import logging
import os
import select
import struct
import tempfile
from urllib.parse import ParseResult, urlparse, urlunparse

import shutil
import socket
from dataclasses import dataclass
from typing import Iterable, Dict, NamedTuple, Type, List, Optional, TypeVar, Generic, Tuple

from xrpc.net import json_pack, json_unpack, RPCPacket
from xrpc.serde.abstract import SerdeStruct
from xrpc.trace import log_tr_net_raw_out, log_tr_net_raw_err, log_tr_net_raw_in, log_tr_net_meta_out, \
    log_tr_net_meta_in, log_tr_net_obj_out, log_tr_net_obj_in, log_tr_net_sel_err, log_tr_net_sel_in, trc
from xrpc.util import _log_called_from

Origin = str

PACKET_PACKER = struct.Struct('!I')


def _norm_addr(addr: Origin):
    pr: ParseResult = urlparse(addr)

    pr = pr._replace(params=None, query=None, fragment=None)

    return urlunparse(pr)


class RPCPacketRaw(NamedTuple):
    addr: Origin
    packet: RPCPacket


@dataclass(eq=True, frozen=True)
class Packet:
    addr: Origin
    data: bytes

    def __repr__(self):
        txt = repr(self.data[:12])

        if len(self.data) > 12:
            txt += '~'

        return f'{self.__class__.__name__}({self.addr}. {txt})'

    def pack(self) -> bytes:
        body = self.data

        len_bts = PACKET_PACKER.pack(len(body))

        return len_bts + body

    @classmethod
    def unpack(cls, addr: Origin, buffer: bytes):
        buffer = memoryview(buffer)

        if len(buffer) < 4:
            raise ValueError('Could not parse size')
        size, = PACKET_PACKER.unpack(buffer[:4])
        if len(buffer) < 4 + size:
            raise ValueError('Size is not enough')

        y = Packet(addr, buffer[4:size + 4].tobytes())

        assert len(buffer) == size + 4, (len(buffer), size + 4)

        return y


class Transport:
    def __init__(self, url):
        self.url: ParseResult = urlparse(url)
        self._fd = None

    def __len__(self):
        return len(self.fds)

    @property
    def fd(self) -> socket.socket:
        assert self._fd, '_fd must be set'
        return self._fd

    @fd.setter
    def fd(self, value):
        self._fd = value

    @property
    def origin(self) -> Origin:
        raise NotImplementedError()

    @property
    def fds(self) -> List[socket.socket]:
        return [self.fd]

    def send(self, packet: Packet):
        raise NotImplementedError()

    def read(self, polled_flags: Optional[List[bool]] = None) -> Iterable[Packet]:
        raise NotImplementedError()

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def from_url(cls, url) -> 'Transport':
        parsed: ParseResult = urlparse(url)

        trc().debug('%s %s', url, parsed)

        if parsed.scheme in TRANSPORT_MAP:
            return TRANSPORT_MAP[parsed.scheme](url)
        else:
            raise ValueError(f'Could not find transport for scheme `{parsed.scheme}`')


class FauxTransport(Transport):
    def __init__(self, *fds: socket.socket):
        self._fds = list(fds)

    @property
    def fds(self) -> List[socket.socket]:
        return self._fds


class Transports(Transport):
    def __init__(self, **kwargs: Transport):
        self.tss: Dict[str, Transport] = kwargs

    def _mapping(self):
        iter_obj = enumerate((ti, fd) for _, (ti, v) in enumerate(sorted(self.tss.items())) for fd in v.fds)
        iter_obj = ((tti, ti, fd) for tti, (ti, fd) in iter_obj)

        return iter_obj

    def poll_helper(self, poll_flags: Optional[List[bool]] = None) -> Dict[str, List[bool]]:
        r = {}

        for tti, ti, fd in self._mapping():
            if ti not in r:
                r[ti] = []

            if poll_flags is None:
                r[ti].append(True)
            else:
                r[ti].append(poll_flags[tti])
        return r

    @property
    def fds(self) -> List[socket.socket]:
        return [x for _, _, x in self._mapping()]

    def __setitem__(self, key, value):
        self.tss[key] = value

    def __getitem__(self, item):
        return self.tss[item]


def recvfrom_helper(fd, buffer_size=2 ** 16, logger_name='net.trace.raw'):
    try:
        while True:
            try:
                buffer, addr = fd.recvfrom(buffer_size)
            except OSError as e:
                if e.errno == 9:
                    raise ConnectionAbortedError(str(e))
                else:
                    raise

            if len(buffer) == 0:
                raise ConnectionAbortedError('Zero bytes received')

            try:
                yield Packet.unpack(addr, buffer)
            except ValueError:
                logging.getLogger(f'{logger_name}.e').error('[%d] %s %s', len(buffer), addr, buffer)
                break
    except BlockingIOError:
        return


class UDPTransport(Transport):
    # transport needs to be able to discern the clients it announces itself

    # we need this on both sides, really
    # Sender and Receiver
    def __init__(self, url, buffer=2 ** 16):
        super().__init__(url)

        assert isinstance(buffer, int), buffer

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP

        try:
            sock.bind((self.url.hostname if self.url.hostname else '0.0.0.0', self.url.port if self.url.port else 0))
        except OSError as e:
            raise OSError(f'{url}')
        sock.settimeout(0)

        self.url = url

        self.fd: socket = sock
        self.buffer = buffer

    def __len__(self):
        return 1

    @property
    def origin(self) -> Origin:
        addr, port = self.fd.getsockname()
        return f'udp://{addr}:{port}'

    def send(self, packet: Packet):
        addr = packet.addr

        if isinstance(addr, str):
            pr: ParseResult = urlparse(addr)
            assert pr.scheme == 'udp' and pr.port is not None, pr

            addr = (pr.hostname, int(pr.port))
        else:
            raise NotImplementedError(repr(addr))

        log_tr_net_raw_out.debug('[%d] %s %s', len(packet.data), addr, packet.data)

        try:
            return self.fd.sendto(packet.pack(), addr)
        except OSError as e:
            if str(e).endswith('Invalid argument'):
                log_tr_net_raw_err.error('%s %s %s', self.url, addr, e)
                return None
        except socket.gaierror as e:
            if str(e).endswith('Try again'):
                log_tr_net_raw_err.debug('%s %s', addr, e)
                return None
            elif str(e).endswith('Name or service not known'):
                log_tr_net_raw_err.debug('%s %s', addr, e)
                return None
            else:
                raise ConnectionError(str(e))

    def read(self, polled_flags: Optional[List[bool]] = None) -> Iterable[Packet]:
        for x in recvfrom_helper(self.fd, self.buffer):
            log_tr_net_raw_in.debug('[%d] %s %s', len(x.data), x.addr,
                                                       x.data)

            addr, port = x.addr

            yield Packet(f'udp://{addr}:{port}', x.data)

    def close(self):
        log_tr_net_meta_out.debug('[3] %s Close %s', self, self.fd)
        self.fd.close()


def _insert_ordered(insert_to, insert_val, insert_item, ord_col, cmp_fn=lambda x: x):
    """

    :param insert_to: ``len(insert_to) = len(ord_col) - 1``
    :param insert_val: value to insert into ``insert_to``
    :param insert_item: value inserted into ``ord_col``
    :param ord_col: collection of ``insert_item`` ordered by cmp_fn
    :param cmp_fn:
    :return:
    """
    idx = ord_col.index(cmp_fn(insert_item))

    return insert_to[:idx] + [insert_val] + insert_to[idx:]


class UnixTransport(UDPTransport):
    def __init__(self, url, buffer_size=2 ** 24):
        # todo implement multi-server client transport by connecting when sending
        # todo how do we handle gc'ing of the conns ?

        Transport.__init__(self, url)

        self.buffer_size = buffer_size

        parsed: ParseResult = urlparse(url)

        should_bind = parsed.fragment == 'bind'

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
        sock.setblocking(False)
        sock.settimeout(0)

        self._fd_bind: Optional[socket] = None
        self._client_ctr = 0
        self._fd_clients: Dict[Origin, socket] = {}

        self._temp_dir = None

        path = parsed.path

        if not path:
            self._temp_dir = tempfile.mkdtemp()

            path = os.path.join(self._temp_dir, 'unix.sock')

        self._path = path

        if should_bind:
            self._fd_bind = sock
            log_tr_net_meta_in.debug('Bind `%s`', self._path)
            sock.bind(self._path)
            sock.listen(1024)

        else:
            self._fd_clients[_norm_addr(url)] = sock
            sock.connect(parsed.path)

            log_tr_net_meta_out.debug('Connect `%s`', parsed.path)

    def __len__(self):
        return len(self._fd_clients) if self._fd_bind else 1

    @property
    def _next_client_id(self):
        r = self._client_ctr
        self._client_ctr += 1
        return r

    @property
    def _fd_clients_ord(self):
        return [(x, y) for x, y in sorted(self._fd_clients.items())]

    @property
    def fds(self) -> List[socket.socket]:
        bind_fd = []
        if self._fd_bind:
            bind_fd = [self._fd_bind]
        r = bind_fd + [x for _, x in self._fd_clients_ord]
        return r

    @property
    def origin(self) -> Origin:
        return f'unix://{self._path}'

    def _clean_client_id(self, addr, reason=None):
        log_tr_net_meta_in.debug('Close %s %s', addr, reason)

        if addr in self._fd_clients:
            cli = self._fd_clients[addr]
            del self._fd_clients[addr]
            cli.close()
        else:
            raise ValueError(f'{addr} {reason}')

    def send(self, packet: Packet):
        addr = packet.addr

        # todo: KeyError if the address had been subsequently removed

        addr = _norm_addr(addr)

        if addr not in self._fd_clients:
            log_tr_net_raw_err.error('Drop %s %s %s', self, addr, packet)
            for x in self._fd_clients.keys():
                log_tr_net_raw_err.getChild('dump').error('%s', x)
            _log_called_from(log_tr_net_raw_err.getChild('tb'))
            return

        sock = self._fd_clients[addr]

        log_tr_net_raw_out.debug('[%d] %s %s', len(packet.data), addr, packet.data)

        try:
            r = sock.send(packet.pack())
            trc('return').debug('%s', r)
            return r
        except BrokenPipeError as e:
            self._clean_client_id(addr, str(e))
            log_tr_net_raw_err.error('[%d] %s %s', len(packet.data), addr, packet.data)
            # todo should we raise or should we ignore the error ?
            # todo upstream possibly needs to know when a client was disconnected
            #raise ConnectionAbortedError(addr, str(e))

    def read(self, polled_flags: Optional[List[bool]] = None) -> Iterable[Packet]:
        if polled_flags is None:
            polled_flags = [True for _ in self._fd_clients_ord]

            if self._fd_bind:
                polled_flags = [True] + polled_flags

        if self._fd_bind:
            polled_bind, *polled_flags = polled_flags
        else:
            polled_bind = False

        if polled_bind:
            while True:
                try:
                    c_sock, _ = self._fd_bind.accept()
                    c_sock.setblocking(False)
                    c_sock.settimeout(0)

                    addr = self._next_client_id

                    c_addr_url = _norm_addr(f'unix://{addr}{self._path}')

                    self._fd_clients[c_addr_url] = c_sock

                    polled_flags = _insert_ordered(polled_flags, True, c_addr_url, [x for x, _ in self._fd_clients_ord])

                    log_tr_net_meta_in.debug('Accept %s %s', c_addr_url, addr)
                except BlockingIOError:
                    break

        flags_clients = list(zip(polled_flags, self._fd_clients_ord))

        for is_ready, (addr, fd) in flags_clients:
            if not is_ready:
                continue

            try:
                for x in recvfrom_helper(fd, self.buffer_size):
                    log_tr_net_raw_in.debug('[%d] %s %s', len(x.data), addr,
                                                               x.data)

                    yield Packet(addr, x.data)
            except ConnectionAbortedError as e:
                if addr not in self._fd_clients:
                    log_tr_net_raw_err.error('%s %s', addr, fd)
                    continue

                self._clean_client_id(addr, str(e))
                # todo should we raise or should we ignore the error ?
                # todo upstream possibly needs to know when a client was disconnected
                raise ConnectionAbortedError(addr, str(e))

    def close(self):
        if self._temp_dir:
            shutil.rmtree(self._temp_dir)
            self._temp_dir = None

        for _, fd in self._fd_clients_ord:
            log_tr_net_meta_in.debug('[1] Close %s', fd)
            fd.close()

        self._fd_clients = {}

        if self._fd_bind:
            log_tr_net_meta_in.debug('[2] Close %s', self._fd_bind)
            self._fd_bind.close()
            self._fd_bind = None


TRANSPORT_MAP: Dict[str, Type[Transport]] = {
    'udp': UDPTransport,
    'unix': UnixTransport,
}

# EVTerminator = Callable[[Any], None]
# EVException = Callable[[Exception], bool]


# @dataclass
# class EVErrEntry:
#    term: EVTerminator
#    exc: EVException


TReq = TypeVar('TReq')
TRep = TypeVar('TRep')


class TransportSerde(Generic[TReq, TRep]):
    def __init__(self, t: Transport, serde: SerdeStruct, cls_req: Type[TReq], cls_rep: Type[TRep]):
        self.transport = t
        self.serde = serde
        self.cls_req = cls_req
        self.cls_rep = cls_rep

    @property
    def fds(self) -> List[socket.socket]:
        return self.transport.fds

    def send(self, to: Origin, obj: TReq):
        pkt_body = json_pack(self.serde.serialize(self.cls_req, obj))

        log_tr_net_obj_out.debug('[%s] %s', self.cls_req, obj)

        pkt = Packet(to, pkt_body)

        self.transport.send(pkt)

    def read(self, polled_flags: Optional[List[bool]] = None) -> Iterable[Tuple[Origin, TRep]]:
        for x in self.transport.read(polled_flags):
            obj = self.serde.deserialize(self.cls_rep, json_unpack(x.data))
            log_tr_net_obj_in.debug('[%s] %s', self.cls_rep, obj)
            yield x.addr, obj


def select_helper(fds, max_wait: Optional[float] = None) -> Optional[List[bool]]:
    log_tr_net_sel_in.debug('%s %s', max_wait, fds)

    try:
        rd, _, er = select.select(fds, [], fds, max_wait)
    except OSError as e:
        log_tr_net_sel_err.debug('%s %s %s', e, max_wait, fds)

        if e.errno == 9:
            raise ConnectionAbortedError(str(e))
        else:
            raise
    except KeyboardInterrupt:
        raise
    except:
        raise ValueError(f'{fds} {max_wait}')

    r = set(rd + er)

    return [fd in r for fd in fds]



import heapq
from collections import deque

from datetime import datetime
from typing import TypeVar, Generic, List, Optional, Dict, Tuple, Callable, Type, Iterator

from xrpc.util import time_now


class MutableInt:
    def __init__(self, state: int = 0):
        self.state = int(state)

    def __set__(self, instance, value):
        pass

    def __iadd__(self, other):
        self.state += other

    def __isub__(self, other):
        self.state -= other

    def __le__(self, other):
        return self.state <= other

    def set(self, state: int):
        self.state = int(state)

    def reduce(self, x=1):
        self.state -= x

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.state})'


class MutableDateTime:
    def __init__(self, t: datetime):
        self.t = t

    @classmethod
    def now(cls):
        return MutableDateTime(time_now())

    def get(self) -> datetime:
        return self.t

    def set(self, t: datetime) -> datetime:
        r = self.t
        self.t = t
        return r


V = TypeVar('T')
K = TypeVar('K')


class Queue(Generic[V]):
    def copy(self) -> 'Queue':
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def push(self, val: V):
        raise NotImplementedError()

    def pop(self) -> V:
        raise NotImplementedError()

    def peek(self) -> Optional[V]:
        raise NotImplementedError()

    def iter(self) -> Iterator[V]:
        raise NotImplementedError()


class BinaryQueue(Queue[V]):
    def __init__(self, initial: Optional[List[V]] = None):
        if initial:
            self.h = deque(sorted(initial))
        else:
            self.h = deque()

    def __len__(self):
        return len(self.h)

    def copy(self) -> 'Queue':
        bq = BinaryQueue()
        bq.h = deque(self.h)
        return bq

    def iter(self) -> Iterator[V]:
        for x in self.h:
            yield x

    def push(self, val: V):
        if len(self.h) == 0:
            self.h.append(val)
            return

        min_idx = 0
        max_idx = len(self.h) - 1

        while min_idx != max_idx and min_idx < max_idx - 1:
            mid_idx = (min_idx + max_idx) // 2

            mid_val = self.h[mid_idx]

            prev_range = (min_idx, max_idx)

            if mid_val == val:
                break
            elif mid_val > val:
                max_idx = mid_idx - 1
            else:
                min_idx = mid_idx + 1

            assert prev_range != (min_idx, max_idx), f'Ranges must always differ, {prev_range}, {(min_idx, max_idx)}'

        if min_idx == max_idx:
            if self.h[min_idx] < val:
                self.h.insert(min_idx + 1, val)
            else:
                self.h.insert(min_idx, val)
            return
        else:
            if self.h[min_idx] > val:
                self.h.insert(min_idx, val)
            elif self.h[max_idx] < val:
                self.h.insert(max_idx + 1, val)
            else:
                self.h.insert(min_idx + 1, val)
            return

    def pop(self) -> V:
        r = self.h.popleft()
        return r

    def peek(self) -> Optional[V]:
        if len(self.h):
            return self.h[0]
        else:
            return None


class HeapQueue(Queue[V]):
    def __init__(self, initial: Optional[List[V]] = None):
        if initial:
            self.h = list(initial)
            heapq.heapify(self.h)
        else:
            self.h = []

    def __len__(self):
        return len(self.h)

    def iter(self) -> Iterator[V]:
        copy = self.copy()

        while True:
            try:
                yield copy.pop()
            except IndexError:
                return

    def copy(self) -> 'Queue':
        hq = HeapQueue()
        hq.h = list(self.h)
        return hq

    def push(self, val: V):
        heapq.heappush(self.h, val)

    def pop(self) -> V:
        return heapq.heappop(self.h)

    def peek(self) -> Optional[V]:
        if len(self.h):
            return self.h[0]
        else:
            return None

    def pushpop(self, val: V):
        return heapq.heappushpop(self.h, val)

    def replace(self, val: V):
        return heapq.heapreplace(self.h, val)


Ord = TypeVar('Ord')


class KeyedQueue(Generic[Ord, K, V], Queue[V]):
    def __init__(self,
                 initial: Optional[List[V]] = None,
                 *,
                 ord: Callable[[V], Ord] = lambda x: x,
                 key: Callable[[V], K] = lambda x: x,
                 q_cls: Type[Queue] = HeapQueue
                 ):
        self.ord_fn = ord
        self.key_fn = key

        self.next_ctr: Dict[K, int] = {}
        self.ref_ctr: Dict[K, int] = {}
        self.values: Dict[K, V] = {}

        self.q_cls = q_cls
        self.q: Queue[Tuple[Ord, K, int]] = q_cls()

        if initial:
            for x in initial:
                self.push(x)

    def get(self, key: K, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __len__(self):
        return len(self.values)

    def __getitem__(self, key: K):
        return self.values[key]

    def __delitem__(self, key: K):
        del self.values[key]

    def __contains__(self, key: K):
        return key in self.values

    def keys(self):
        return self.values.keys()

    def iter(self) -> Iterator[V]:
        for _, key, idx in self.q.iter():
            if not self._pop_checks(key, idx):
                continue

            yield self.values[key]

    def copy(self):
        khq = KeyedQueue(ord=self.ord_fn, key=self.key_fn, q_cls=self.q_cls)
        khq.next_ctr = dict(self.next_ctr)
        khq.values = dict(self.values)
        khq.q = self.q.copy()
        return khq

    def push(self, val: V):
        key = self.key_fn(val)
        ord_ = self.ord_fn(val)

        if key not in self.next_ctr:
            self.next_ctr[key] = 0
            self.ref_ctr[key] = 0

        idx = self.next_ctr[key]
        self.next_ctr[key] += 1
        self.ref_ctr[key] += 1

        self.values[key] = val

        self.q.push((ord_, key, idx))

    def _pop_checks(self, key, idx):
        if key not in self.values:
            return False

        # this will not work, if a newer item had been subsequently added

        if self.next_ctr[key] - 1 != idx:
            return False

        return True

    def pop(self) -> V:
        while True:
            ord_, key, idx = self.q.pop()

            if not self._pop_checks(key, idx):
                continue

            self.ref_ctr[key] -= 1

            if self.ref_ctr[key] == 0:
                if key in self.next_ctr:
                    del self.next_ctr[key]
                del self.ref_ctr[key]
            return_ = self.values[key]
            del self.values[key]

            return return_

    def peek(self) -> Optional[V]:
        while True:
            val = self.q.peek()

            if val is None:
                return None

            ord_, key, idx = val

            if key not in self.values:
                self.q.pop()
                continue

            if self.next_ctr[key] - 1 != idx:
                self.q.pop()
                continue

            return self.values[key]



import logging
import os
import random
import traceback
from collections import deque
from contextlib import ExitStack
from inspect import getfullargspec, ismethod

import threading
import types
from argparse import ArgumentParser
from dataclasses import dataclass, fields
from datetime import datetime, timedelta
from signal import SIGKILL, SIGTERM, SIGINT
from typing import Callable, Optional, Dict, Deque, TypeVar, Generic, Type, Tuple, Union, List, Any, Iterable, Set

from xrpc.abstract import KeyedQueue
from xrpc.actor import run_server
from xrpc.cli import Parsable
from xrpc.client import ClientConfig, build_wrapper
from xrpc.dsl import rpc, RPCType, regular, signal, DEFAULT_GROUP
from xrpc.error import TimeoutError, TerminationException
from xrpc.logging import logging_config, LoggerSetup, logging_setup, circuitbreaker, cli_main, logging_parser
from xrpc.loop import EventLoop
from xrpc.net import RPCKey
from xrpc.popen import popen
from xrpc.runtime import service, sender, origin, reset, context
from xrpc.serde.types import build_types, ARGS_RET, PairSpec
from xrpc.service import ServiceDefn
from xrpc.trace import trc
from xrpc.transport import Origin, Transport
from xrpc.util import time_now, signal_context


@dataclass
class ClusterConf(Parsable):
    heartbeat: float = 5.
    max_pings: int = 5
    metrics: float = 10.
    retry_delta: float = 0.10
    pending_max: int = 10


@dataclass
class WorkerConf(Parsable):
    processes: int = 3
    threads: int = 7

    @property
    def total(self):
        return self.processes * self.threads


@dataclass
class BrokerConf(Parsable):
    backlog: int = 100
    """Maximum number of unassigned items in the queue"""
    running_backlog: Optional[int] = None
    """How many maximum running tasks are allowed to be"""
    flush_backlog: int = 100
    """Maximum number of items that have not been passed as a result upstream"""


@dataclass
class WorkerMetric:
    brid: Optional[RPCKey]
    jobs: int
    pending: int
    workers: int
    workers_free: int
    broker_url: str


@dataclass
class BrokerMetric:
    workers: int
    capacity: int
    jobs_pending: int
    jobs: int
    assigned: int

    resigns: int
    dones: int

    @property
    def running(self) -> int:
        return self.jobs_pending + self.assigned

    @property
    def flushing(self) -> int:
        return self.jobs - self.running


NodeMetric = Union[WorkerMetric, BrokerMetric]

RequestType = TypeVar('RequestType', Any, str)
ResponseType = TypeVar('ResponseType', Any, str)

WorkerCallable = Callable[[RequestType], ResponseType]


class MetricCollector:
    @rpc(RPCType.Signalling)
    def metrics(self, metric: NodeMetric):
        pass


def get_func_types(fn: WorkerCallable) -> Tuple[Type[RequestType], Type[ResponseType]]:
    if not isinstance(fn, types.FunctionType):
        fn = fn.__call__

    spec = getfullargspec(fn)
    is_method = ismethod(fn)
    annot = build_types(spec, is_method, allow_missing=True)
    arg = next(PairSpec(spec, is_method)(None))

    return annot[arg.name], annot[ARGS_RET]


def worker_inst(logger_config: LoggerSetup, fn: WorkerCallable, unix_url: str):
    def sigint_handler(code, frame):
        stack_list = traceback.StackSummary.extract(
            traceback.walk_stack(frame),
            limit=None,
            capture_locals=True,
        )
        stack_list.reverse()

        for f, x in zip(stack_list, traceback.format_list(stack_list)):
            locs = None

            if hasattr(f, 'locals'):
                locs = f.locals

            xs = x.split('\n')
            if len(xs) and xs[-1] == '':
                xs = xs[:-1]
            for y in xs:
                logging.getLogger('traceback').error('%s', y)

            # logging.getLogger('traceback.locals').error('%s', locs)

    with ExitStack() as es:
        stacks = [
            logging_setup(logger_config),
            circuitbreaker(main_logger='broker'),
            signal_context(signals=(SIGINT,), handler=sigint_handler)
        ]

        for stack in stacks:
            es.enter_context(stack)

        logging.getLogger('worker_inst').info(f'Start %s', unix_url)

        # use the callable's type hints in order to serialize and deserialize parameters

        cls_req, cls_res = get_func_types(fn)

        should_inst = False

        try:
            run_server(WorkerForkInst, WorkerForkInst(), {DEFAULT_GROUP: unix_url})
        except ForkException:
            should_inst = True

        if should_inst:
            run_server(WorkerInst[cls_req, cls_res], WorkerInst(fn), {
                DEFAULT_GROUP: unix_url,
                BACKEND: 'unix://#bind'
            })


class ForkException(BaseException):
    pass


class WorkerForkInst:
    @rpc(RPCType.Signalling)
    def fork(self):
        trc('0').error('%s', sender())
        if os.fork() == 0:
            context().actor.get('signal')._fork_transport_patch()
            raise ForkException()

    @rpc(exc=True)
    def exc(self, exc: BaseException) -> bool:
        trc('0').error('%s', repr(exc))
        return False

    @signal()
    def exit(self):
        raise TerminationException()

    @regular()
    def announce(self) -> Optional[float]:
        s = service(WorkerAnnounce, origin())
        s.bk_announce(True, os.getpid())

        return None


def worker_thread_main(fn, unix_url: str):
    cls_req, cls_res = get_func_types(fn)

    run_server(WorkerThreadInst[cls_req, cls_res], WorkerThreadInst(fn), {
        DEFAULT_GROUP: unix_url
    })


class WorkerThreadInst(Generic[RequestType, ResponseType]):
    def __init__(self, fn: WorkerCallable):
        self.fn = fn
        self.cls_req, self.cls_res = get_func_types(self.fn)

    @regular()
    def init(self) -> Optional[float]:
        x: WorkerInst[self.cls_req, self.cls_res] = service(WorkerInst[self.cls_req, self.cls_res], origin())
        x.bk_started()
        return None

    @rpc(RPCType.Signalling)
    def put(self, payload: RequestType):
        ret = self.fn(payload)

        x: WorkerInst[self.cls_req, self.cls_res] = service(WorkerInst[self.cls_req, self.cls_res], origin())
        x.bk_done(ret)


BACKEND = 'backend'


class WorkerInst(Generic[RequestType, ResponseType]):
    def __init__(self, fn: WorkerCallable):
        self.fn = fn
        self.cls_req, self.cls_res = get_func_types(self.fn)
        self.started = False
        self.thread_addr: Optional[str] = None
        self.thread = None

    @rpc(RPCType.Signalling)
    def put(self, payload: Optional[RequestType]):
        s: WorkerThreadInst[self.cls_req, self.cls_res] = service(
            WorkerThreadInst[self.cls_req, self.cls_res],
            self.thread_addr,
            group=BACKEND,
        )
        s.put(payload)

    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_started(self):
        self.thread_addr = sender()
        reset(self.check_started, None)
        reset(self.announce, 0)

    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_done(self, ret: ResponseType):
        s = service(Worker[self.cls_req, self.cls_res], origin(DEFAULT_GROUP), group=DEFAULT_GROUP)
        s.bk_done(ret)

    @rpc(exc=True, group=BACKEND)
    def bk_exc(self, exc: ConnectionAbortedError) -> bool:
        trc().error('Lost connection to the worker thread')
        raise TerminationException()

    @rpc(exc=True)
    def exc(self, exc: ConnectionAbortedError) -> bool:
        raise TerminationException()

    @regular()
    def start_bg(self):
        self.thread = threading.Thread(target=worker_thread_main, args=(self.fn, origin(BACKEND)))
        self.thread.daemon = True
        self.thread.start()
        self.started = True

    @regular()
    def announce(self) -> Optional[float]:
        s = service(WorkerAnnounce, origin())
        s.bk_announce(False, os.getpid())

        return None

    @regular(initial=5)
    def check_started(self):
        raise TerminationException()

    @signal(codes=(SIGTERM,))
    def exit(self):
        raise TerminationException()


@dataclass(frozen=False)
class WorkerLoad:
    occupied: int = 0
    capacity: int = 0


SchedKeyQueue = KeyedQueue[datetime, RPCKey, 'SchedKey']


@dataclass(frozen=True)
class SchedKey:
    when: datetime
    key: RPCKey

    @classmethod
    def now(cls, key: RPCKey):
        return SchedKey(time_now(), key)

    @classmethod
    def queue(cls) -> SchedKeyQueue:
        return KeyedQueue(ord=SchedKey.ord_fn,
                          key=SchedKey.key_fn)

    @classmethod
    def ord_fn(cls, x: 'SchedKey'):
        return x.when

    @classmethod
    def key_fn(cls, x: 'SchedKey'):
        return x.key


def process_queue(
        jobs_pending_done: SchedKeyQueue,
        fn: Callable[[RPCKey], Optional[float]],
        tol: float = 0.
) -> Optional[float]:
    # sb = service(Broker[self.cls_req, self.cls_res], self.url_broker, group=BACKEND)

    now = time_now()

    while True:
        val = jobs_pending_done.peek()

        if val is None:
            return None

        dsec = (val.when - now).total_seconds()

        if dsec > tol:
            return dsec

        val = jobs_pending_done.pop()

        next_timeout = fn(val.key)

        if next_timeout:
            jobs_pending_done.push(SchedKey(now + timedelta(seconds=next_timeout), val.key))


class WorkerAnnounce:
    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_announce(self, is_fork: bool, wpid: int):
        raise NotImplementedError()


class Worker(Generic[RequestType, ResponseType], WorkerAnnounce):
    def __init__(
            self,
            cls_req: Type[RequestType], cls_res: Type[ResponseType],
            conf: ClusterConf,
            url_broker: Origin,
            fn: WorkerCallable[RequestType, ResponseType],
            url_metrics: Optional[str] = None,
            par_conf: WorkerConf = WorkerConf(),
    ):
        self.brid: Optional[RPCKey] = None
        self.has_started = False
        self.has_terminated = False
        self.par_conf = par_conf

        self.load = WorkerLoad()

        self.cls_req = cls_req
        self.cls_res = cls_res

        self.conf = conf
        self.url_broker = url_broker
        self.url_metrics = url_metrics

        self.fn = fn

        # todo: a worker A with PID 1 dies. Then we create a new worker B and it gets the same PID
        # todo: we still don't know that A is dead but already know about B
        # todo: we then realise A had died, so we try to evict it, instead killing B
        self.workers_fork_addrs: Dict[str, int] = {}
        self.worker_addrs: Dict[str, int] = {}

        self.jobs: Dict[RPCKey, RequestType] = {}
        self.jobs_res: Dict[RPCKey, ResponseType] = {}
        self.jobs_workers: Dict[RPCKey, str] = {}
        self.workers_jobs: Dict[str, RPCKey] = {}
        self.workers_free: KeyedQueue[str, str, str] = KeyedQueue()

        self.workers_fork: Deque[str] = deque()

        self.jobs_pending_done: SchedKeyQueue = SchedKey.queue()

    def _start_worker_inst(self):
        popen(worker_inst, logging_config(), self.fn, origin(BACKEND))

    @regular()
    def ep(self) -> Optional[float]:
        for y in range(self.par_conf.processes):
            self._start_worker_inst()
        return None

    def _free(self, jid: RPCKey) -> Optional[str]:
        if jid not in self.jobs_workers:
            return None

        wid = self.jobs_workers[jid]
        del self.jobs_workers[jid]
        del self.workers_jobs[wid]
        return wid

    def _evict(self, jid: RPCKey):
        del self.jobs[jid]
        wid = self._free(jid)

        if jid in self.jobs_res:
            del self.jobs_res[jid]

        if wid is not None:
            if wid in self.workers_free:
                del self.workers_free[wid]

            wpid = self.worker_addrs[wid]
            del self.worker_addrs[wid]

            self.load.capacity -= 1
            self.load.occupied -= 1

            os.kill(wpid, SIGKILL)

    def _fork(self):
        sel_wfid = random.choice(list(self.workers_fork_addrs.keys()))
        self.workers_fork.append(sel_wfid)
        reset(self.push_fork, 0)

    def _evict_all(self, terminating=False):
        for jid in list(self.jobs.keys()):
            self._evict(jid)

            if not terminating:
                self._fork()

    def _brid_new(self, brid: RPCKey):
        should_reset = self.brid is not None and self.brid != brid
        self.brid = brid

        # todo when an brid change happens
        # todo there's a chance the jid will not be assigned to any workers
        # todo yet it's gointg to be listed in jobs.

        if should_reset:
            self._evict_all()

    @rpc(RPCType.Signalling)
    def assign(self, brid: RPCKey, jid: RPCKey, jreq: RequestType):
        # todo: currently sends to an address will cause an exception if we're lucky enough to
        # todo: send a task right after the worker had died (and before transport realised that)

        sb = service(Broker[self.cls_req, self.cls_res], self.url_broker)

        if brid != self.brid:
            trc('brid').error('%s != %s %s', brid, self.brid, jid)
            reset(self.push_announce, 0)
            sb.bk_assign(brid, jid, False)
            return

        if jid in self.jobs:
            # is there any scenario where a job may be assigned to something else ?
            trc('kno').error('%s', jid)
            sb.bk_assign(brid, jid, True)
            return

        if len(self.workers_free) == 0 or len(self.jobs_res) >= self.conf.pending_max:
            sb.bk_assign(brid, jid, False)
            return

        nwid = self.workers_free.pop()

        self.jobs[jid] = jreq
        self.jobs_workers[jid] = nwid
        self.workers_jobs[nwid] = jid

        s = service(WorkerInst[self.cls_req, self.cls_res], nwid, group=BACKEND)
        s.put(jreq)

        self.load.occupied += 1

        sb.bk_assign(brid, jid, True)

    @rpc(RPCType.Signalling)
    def resign(self, brid: RPCKey, jid: RPCKey, reason: Optional[str] = None):
        if brid != self.brid:
            trc('brid').error('%s != %s %s', brid, self.brid, jid)
            reset(self.push_announce, 0)
            return

        self._push_done(jid)

        if jid not in self.jobs:
            # [w-1] resignation notice may appear after worker had successfully finished the job
            # [w-1] in such a scenario, a broker must report resignation as a failure by checking it's finish log
            trc('unk').error('%s', jid)
            return

        if jid in self.jobs_res:
            trc('done').error('%s', jid)
            return

        self._evict(jid)
        self._fork()
        return

    @rpc(RPCType.Signalling)
    def registered(self, brid: RPCKey):
        if self.brid != brid:
            self._brid_new(brid)

    @rpc(RPCType.Signalling)
    def done_ack(self, brid: RPCKey, jid: RPCKey):
        if brid != self.brid:
            trc('brid').error('%s != %s %s', brid, self.brid, jid)
            reset(self.push_announce, 0)
            return

        if jid in self.jobs_res:
            self._evict(jid)

            if jid in self.jobs_pending_done:
                del self.jobs_pending_done[jid]

    def _push_done(self, jid: RPCKey):
        sb = service(Broker[self.cls_req, self.cls_res], self.url_broker)

        if jid in self.jobs:
            if jid in self.jobs_res:
                sb.bk_done(self.brid, jid, True, self.jobs_res[jid])
            else:
                sb.bk_done(self.brid, jid, False)

        return self.conf.retry_delta

    @regular()
    def push_done(self) -> Optional[float]:
        return process_queue(
            self.jobs_pending_done,
            self._push_done,
            self.conf.retry_delta * 0.1
        )

    @regular()
    def push_fork(self) -> Optional[float]:
        while len(self.workers_fork):
            wid = self.workers_fork.popleft()

            if self.load.capacity >= self.par_conf.total:
                continue

            s = service(WorkerForkInst, wid, group=BACKEND)
            # this specific call _may_ cause the event loop to re-loop.
            s.fork()
        return None

    @regular(initial=None)
    def push_announce(self) -> float:
        s = service(Broker[self.cls_req, self.cls_res], self.url_broker, group=DEFAULT_GROUP)
        s.bk_announce(self.brid, self.load)

        return self.conf.heartbeat

    @rpc(RPCType.Repliable)
    def metrics(self) -> WorkerMetric:
        return WorkerMetric(
            self.brid,
            len(self.jobs),
            len(self.jobs_pending_done),
            len(self.workers_free) + len(self.workers_jobs),
            len(self.workers_free),
            self.url_broker,
        )

    @regular(initial=None)
    def push_metrics(self) -> float:
        # todo worker should not expose any of the jobs running to the metrics, this is a broker's job
        if self.url_metrics:
            s = service(MetricCollector, self.url_metrics)
            s.metrics(self.metrics())
        return self.conf.metrics if self.url_metrics else None

    @rpc(exc=True, group=BACKEND)
    def bk_exc(self, exc: ConnectionAbortedError) -> bool:
        # one of the workers had disconnected

        host, reason = exc.args
        trc('1').info('%s %s', host, reason)

        if host in self.workers_jobs:
            jid = self.workers_jobs[host]

            self.resign(self.brid, jid)
        elif host in self.workers_fork_addrs and not self.has_terminated:
            raise ValueError(f'WorkerInst `{host}` had died')

        return True

    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_done(self, res: ResponseType):
        wid = sender()

        if wid not in self.workers_jobs:
            trc('1').error('%s', wid)
            return

        jid = self.workers_jobs[wid]

        if jid not in self.jobs_workers:
            trc('1').error('%s', jid)
            return

        self.jobs_res[jid] = res
        self.jobs_pending_done.push(SchedKey(time_now(), jid))
        reset(self.push_done, 0)

        wid = self.jobs_workers[jid]
        del self.jobs_workers[jid]
        del self.workers_jobs[wid]

        self.load.occupied -= 1
        self.workers_free.push(wid)

    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_announce(self, is_fork: bool, wpid: int):
        sdr = sender()

        trc('0').error('%s %s %s', sdr, is_fork, wpid)

        if is_fork:
            if sdr in self.workers_fork_addrs:
                return
            else:
                self.workers_fork_addrs[sdr] = wpid

                s = service(WorkerForkInst, sdr, group=BACKEND)

                for _ in range(self.par_conf.threads):
                    trc('01').error('%s', _)
                    s.fork()
            return

        if sdr in self.worker_addrs:
            return

        trc('1').debug('%s', sdr)

        self.worker_addrs[sdr] = wpid
        self.workers_free.push(sdr)
        self.load.capacity += 1

        if self.load.capacity == self.par_conf.processes * self.par_conf.threads:
            trc('capa').debug('capacity reached')
            self._started()

        if self.has_started:
            self.push_announce()

    def _started(self):
        self.has_started = True

        reset(self.startup_timeout, None)
        reset(self.push_announce, 0)

        if self.url_metrics:
            reset(self.push_metrics, 0)

    @regular(initial=10.)
    def startup_timeout(self):
        assert False, f'Could not reach the required capacity ' \
                      f'{self.load.capacity}/{self.par_conf.processes * self.par_conf.threads}'

    @signal()
    def exit(self):
        self.has_terminated = True

        for wfpid in self.workers_fork_addrs.values():
            os.kill(wfpid, SIGTERM)

        while len(self.workers_free):
            wid = self.workers_free.pop()
            wpid = self.worker_addrs[wid]
            del self.worker_addrs[wid]

            os.kill(wpid, SIGKILL)

        self._evict_all(terminating=True)

        try:
            s = service(Broker[self.cls_req, self.cls_res], self.url_broker, ClientConfig(timeout_total=1.))
            s.bk_unregister()
        except TimeoutError:
            logging.getLogger('exit').error('Could not contact broker')

        raise TerminationException()


@dataclass(frozen=False)
class WorkerState:
    pings_remaining: int
    load: WorkerLoad


@dataclass(frozen=False)
class JobState(Generic[RequestType, ResponseType]):
    req: RequestType
    created: datetime
    attempts: int = 0
    started: Optional[datetime] = None
    res: Optional[ResponseType] = None
    finished: Optional[datetime] = None


class BrokerResult(Generic[ResponseType]):
    @staticmethod
    def _finished_ack(jid: RPCKey):
        s = service(BrokerFlushAck, sender(), group=DEFAULT_GROUP)
        s.flush_ack(jid)

    @rpc(RPCType.Signalling)
    def finished(self, jid: RPCKey, jres: ResponseType):
        logging.getLogger('finished').warning('unused %s', jres)

        self._finished_ack(jid)


class BrokerFlushAck:
    @rpc(RPCType.Signalling)
    def flush_ack(self, jid: RPCKey):
        raise NotImplementedError()


class BrokerEntry(Generic[ResponseType]):
    @rpc(RPCType.Repliable)
    def assign(self, id_: RPCKey, pars: RequestType) -> bool:
        pass


class Broker(Generic[RequestType, ResponseType], BrokerEntry[ResponseType], BrokerFlushAck):
    def __init__(
            self,
            cls_req: Type[RequestType],
            cls_res: Type[ResponseType],
            conf: ClusterConf,
            url_results: Optional[str] = None,
            url_metrics: Optional[str] = None,
            par_conf: BrokerConf = BrokerConf(),
    ):
        self.cls_req = cls_req
        self.cls_res = cls_res

        self.conf = conf
        self.par_conf = par_conf
        self.url_results = url_results
        self.url_metrics = url_metrics

        self.workers: Dict[Origin, WorkerState] = {}
        self.workers_brids: Dict[Origin, RPCKey] = {}

        self.jobs: Dict[RPCKey, JobState[RequestType, ResponseType]] = {}
        self.jobs_pending: KeyedQueue[RPCKey, RPCKey, RPCKey] = KeyedQueue()

        self.jobs_workers: Dict[RPCKey, Origin] = {}
        self.workers_jobs: Dict[Origin, List[RPCKey]] = {}

        self.jobs_pending_assign: SchedKeyQueue = SchedKey.queue()
        self.jobs_pending_resign: SchedKeyQueue = SchedKey.queue()
        self.jobs_pending_flush: SchedKeyQueue = SchedKey.queue()

        self.jobs_cancel: Set[RPCKey] = set()

        self.resigns = 0
        self.dones = 0

        # todo transform broker to a ticket-based system instead

    def _job_new(self, jid: RPCKey, jreq: RequestType):
        trc('0').debug('%s %s', jid, jreq)

        self.jobs[jid] = JobState(jreq, time_now())
        self.jobs_pending.push(jid)

        reset(self.push_push_assign, 0)

    def _job_clean(self, jid: RPCKey) -> str:
        trc('0').error('%s', jid)
        wid = self.jobs_workers[jid]
        del self.jobs_workers[jid]
        self.workers_jobs[wid].remove(jid)
        reset(self.push_push_assign, 0)
        return wid

    def _job_resign(self, jid: RPCKey):
        self.resigns += 1

        self._job_clean(jid)

        if jid in self.jobs_pending:
            del self.jobs_pending[jid]
        if jid in self.jobs_pending_assign:
            del self.jobs_pending_assign[jid]
        if jid in self.jobs_pending_resign:
            del self.jobs_pending_resign[jid]

        if jid in self.jobs_cancel:
            del self.jobs[jid]
            self.jobs_cancel.remove(jid)
        else:
            self.jobs_pending.push(jid)

            self.jobs[jid].started = None
            reset(self.push_push_assign, 0)

    def _job_done(self, jid: RPCKey, jres: ResponseType):
        self.dones += 1

        self._job_clean(jid)
        self.jobs[jid].finished = time_now()
        self.jobs[jid].res = jres

        self.jobs_pending_flush.push(SchedKey.now(jid))
        reset(self.push_flush, 0)

    def _job_flush(self, jid: RPCKey):
        del self.jobs[jid]

        if jid in self.jobs_pending_flush:
            del self.jobs_pending_flush[jid]

    def _push_assign(self, jid: RPCKey) -> Optional[float]:
        wid = self.jobs_workers[jid]
        brid = self.workers_brids[wid]
        s = service(Worker[self.cls_req, self.cls_res], wid)
        s.assign(brid, jid, self.jobs[jid].req)

        return self.conf.retry_delta

    @regular(initial=None)
    def push_assign(self):
        return process_queue(
            self.jobs_pending_assign,
            self._push_assign,
            self.conf.retry_delta * 0.1,
        )

    def _push_resign(self, jid: RPCKey) -> Optional[float]:
        wid = self.jobs_workers[jid]
        brid = self.workers_brids[wid]
        s = service(Worker[self.cls_req, self.cls_res], wid)
        s.resign(brid, jid)

        return self.conf.retry_delta

    @regular(initial=None)
    def push_resign(self):
        return process_queue(
            self.jobs_pending_resign,
            self._push_resign,
            self.conf.retry_delta * 0.1,
        )

    def _push_flush(self, jid: RPCKey) -> Optional[float]:
        if not self.url_results:
            self.flush_ack(jid)
        else:
            assert self.jobs[jid].finished
            s = service(BrokerResult[self.cls_res], self.url_results, group=DEFAULT_GROUP)
            s: BrokerResult[ResponseType]
            s.finished(jid, self.jobs[jid].res)
            return self.conf.retry_delta

    @regular(initial=None)
    def push_flush(self):
        return process_queue(
            self.jobs_pending_flush,
            self._push_flush,
            self.conf.retry_delta * 0.1,
        )

    @rpc(RPCType.Signalling)
    def flush_ack(self, jid: RPCKey):
        if jid not in self.jobs:
            return

        self._job_flush(jid)

    @regular(initial=None)
    def push_push_assign(self):
        def spread(iter_obj: Iterable[Tuple[str, int]]):
            for wid, capa in iter_obj:
                for i in range(capa):
                    yield wid

        def eat(obj: Deque[RPCKey]):
            while len(obj):
                try:
                    yield obj.pop()
                except IndexError:
                    return

        w_caps = (
            (wid, max(0, self.workers[wid].load.capacity - len(self.workers_jobs[wid])))
            for wid, wst
            in self.workers.items()
        )

        jobs_workers = zip(spread(w_caps), eat(self.jobs_pending))

        for wid, jid in jobs_workers:
            trc('1').debug('%s %s', wid, jid)

            self.jobs[jid].started = time_now()
            self.jobs[jid].attempts += 1

            self.jobs_workers[jid] = wid
            self.workers_jobs[wid].append(jid)

            self.jobs_pending_assign.push(SchedKey.now(jid))
            reset(self.push_assign, 0)

    @rpc(RPCType.Repliable)
    def query(
            self,
            since: Optional[RPCKey] = None,
            until: Optional[RPCKey] = None,
            order_by: Optional[Tuple[str, bool]] = None,
            limit: Optional[int] = None
    ) -> Optional[List[Tuple[RPCKey, JobState[RequestType, ResponseType]]]]:
        items = self.jobs.items()

        if order_by:
            order_by_field, order_by_forward = order_by

            order_by_field_defn = [x for x in fields(JobState) if x.name == order_by_field]

            if not len(order_by_field_defn):
                return None

            items = [x for x in items if getattr(x[1], order_by_field) is not None]

            items = sorted(items, key=lambda x: getattr(x[1], order_by_field))

        if limit:
            items = items[:limit]

        r = []

        for k, v in items:
            if since is not None and k < since:
                continue

            if until is not None and k > until:
                continue

            r.append((k, v))

        return r

    def _brid_check(self, brid: RPCKey):
        wid = sender()
        if self.workers_brids.get(wid) != brid:
            trc('brid', depth=2).error('%s', brid)
            return True
        return False

    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_assign(self, brid: RPCKey, jid: RPCKey, ok: bool):
        sdr = sender()
        if sdr not in self.workers:
            trc('3').error('%s', sdr)
            return

        if self._brid_check(brid):
            return

        if jid not in self.jobs:
            trc('1').error('%s', jid)
            return

        if jid not in self.jobs_workers:
            trc('2').error('%s', jid)
            return

        if jid not in self.workers_jobs[sdr]:
            trc('5').error('%s', jid)
            return

        if jid in self.jobs_pending_assign:
            del self.jobs_pending_assign[jid]

        self._worker_pings(brid, sdr)

        if ok:
            return
        else:
            if jid in self.jobs_pending_resign:
                del self.jobs_pending_resign[jid]

            self._job_resign(jid)

    @rpc(RPCType.Signalling, group=BACKEND)
    def bk_done(self, brid: RPCKey, jid: RPCKey, ok: bool, res: Optional[ResponseType] = None):
        w = service(Worker[self.cls_req, self.cls_res], sender())
        w.done_ack(brid, jid)

        if self._brid_check(brid):
            return

        if jid not in self.jobs:
            trc('1').error('%s', jid)
            return

        if self.jobs[jid].finished:
            trc('2').warning('%s', jid)
            return

        if jid in self.jobs_pending_assign:
            del self.jobs_pending_assign[jid]

        if jid in self.jobs_pending_resign:
            del self.jobs_pending_resign[jid]

        if jid not in self.jobs_workers or sender() != self.jobs_workers[jid]:
            return

        self._worker_pings(brid, sender())

        if ok:
            self._job_done(jid, res)
        else:
            self._job_resign(jid)

    @rpc()
    def metrics(self) -> BrokerMetric:
        return BrokerMetric(
            len(self.workers),
            sum(x.load.capacity for x in self.workers.values()),
            len(self.jobs_pending),
            len(self.jobs),
            len(self.jobs_workers),
            self.resigns,
            self.dones
        )

    @regular(initial=None)
    def push_metrics(self) -> float:
        s = service(MetricCollector, self.url_metrics)
        s.metrics(self.metrics())

        return self.conf.metrics

    @rpc(RPCType.Repliable)
    def assign(self, jid: RPCKey, jreq: RequestType) -> bool:
        sp = len(self.jobs_pending)

        if sp > 0 and sp >= self.par_conf.backlog:
            trc('pending').debug('%s', jid)
            return False

        sr = len(self.jobs_workers)

        if self.par_conf.running_backlog and sr > 0 and sr >= self.par_conf.running_backlog:
            trc('running').debug('%s', jid)
            return False

        sf = len(self.jobs_pending_flush)
        if sf > 0 and sf >= self.par_conf.flush_backlog:
            trc('flush').debug('%s', jid)
            return False

        if jid not in self.jobs:
            self._job_new(jid, jreq)

        return True

    @rpc(RPCType.Repliable)
    def cancel(self, jid: RPCKey) -> bool:
        if jid not in self.jobs:
            return False

        if jid in self.jobs_workers:
            self.jobs_cancel.add(jid)
            return self.resign(jid)
        else:
            self._job_flush(jid)

            if jid in self.jobs_pending:
                del self.jobs_pending[jid]

        return True

    @rpc(RPCType.Repliable)
    def resign(self, jid: RPCKey, reason: Optional[str] = None) -> bool:
        if jid not in self.jobs:
            trc('0').debug('%s %s', jid, reason)
            return False

        if jid not in self.jobs_workers:
            trc('1').debug('%s %s', jid, reason)
            return False

        self.jobs_pending_resign.push(SchedKey.now(jid))
        reset(self.push_resign, 0)

        return True

    def _worker_conf_changed(self):
        if self.url_metrics:
            reset(self.push_metrics, 0)
        reset(self.push_push_assign, 0)

    def _worker_new(self, wbrid: RPCKey, wid: Origin, load: WorkerLoad):
        trc().debug('%s %s %s', wbrid, wid, load)

        self.workers[wid] = WorkerState(self.conf.max_pings, WorkerLoad(occupied=0, capacity=load.capacity))
        self.workers_jobs[wid] = []
        self.workers_brids[wid] = wbrid

        self._worker_conf_changed()

    def _worker_lost(self, wid: Origin):
        trc().debug('%s', wid)

        for jid in list(self.workers_jobs[wid]):
            self._job_resign(jid)

        del self.workers[wid]
        del self.workers_jobs[wid]
        del self.workers_brids[wid]

        self._worker_conf_changed()

    @rpc(group=BACKEND)
    def bk_register(self, load: WorkerLoad) -> Optional[RPCKey]:
        wid = sender()

        if wid in self.workers:
            # a worker needs to be unregistered first
            return None

        wbrid = RPCKey.new()

        self._worker_new(wbrid, wid, load)

    @rpc(RPCType.Durable, group=BACKEND)
    def bk_unregister(self):
        wid = sender()

        if wid in self.workers:
            self._worker_lost(wid)

    def _worker_pings(self, brid: RPCKey, wid: str):
        if wid not in self.workers:
            return

        if self.workers_brids[wid] != brid:
            return

        wst = self.workers[wid]

        wst.pings_remaining = self.conf.max_pings

    @rpc(type=RPCType.Signalling, group=BACKEND)
    def bk_announce(self, wbrid: Optional[RPCKey], load: WorkerLoad):
        wid = sender()

        trc('1').debug('%s', wid)

        if wid not in self.workers:
            self._worker_new(RPCKey.new(), wid, load)

        brid = self.workers_brids[wid]

        if wbrid != brid:
            s = service(Worker[self.cls_req, self.cls_res], wid, ClientConfig(timeout_total=1))
            s.registered(self.workers_brids[wid])

        # todo only ping workers through announce

        wst = self.workers[wid]

        flag_changed = wst.load.capacity != load.capacity

        wst.load.capacity = load.capacity
        self._worker_pings(brid, wid)

        if flag_changed:
            self._worker_conf_changed()

    @regular()
    def gc(self) -> float:
        for k in list(self.workers.keys()):
            self.workers[k].pings_remaining -= 1

            trc().debug('%s %s', k, self.workers[k])

            if self.workers[k].pings_remaining <= 0:
                self._worker_lost(k)

        return self.conf.heartbeat

    @signal()
    def exit(self):
        raise TerminationException()

    @regular()
    def startup(self):
        if self.url_metrics:
            reset(self.push_metrics, 0)

        return None


def main(server_url,
         conf=ClientConfig(timeout_total=5),
         **kwargs):
    service_type = Broker[str, str]
    T: Type[service_type] = service_type.__class__

    t = Transport.from_url('udp://0.0.0.0')

    with t:
        ts = EventLoop()
        ets = ts.transport_add(t)
        pt = ServiceDefn.from_cls(service_type)
        r: T = build_wrapper(pt, ets, server_url, conf=conf)

        print(r.metrics())


def parser():
    parser = ArgumentParser()

    logging_parser(parser)

    parser.add_argument(
        '-A'
        '--address',
        dest='server_url',
        default='udp://127.0.0.1:2345'
    )

    return parser


if __name__ == '__main__':
    cli_main(main, parser())



import threading
from typing import Optional, Callable, Dict, TypeVar, Type, Any, List, Union
from urllib.parse import urlparse, ParseResult, urlunparse, urlencode, parse_qsl

from dataclasses import dataclass

import xrpc
from xrpc.client import ClientConfig, ServiceWrapper
from xrpc.loop import EventLoop
from xrpc.service import ServiceDefn
from xrpc.trace import log_tr_exec_in, log_tr_exec_out, trc
from xrpc.transport import Origin
from xrpc.util import time_now

RUNTIME_TL = threading.local()
CTX_NAME = 'rpc_config'
CACHE_NAME = 'rpc_cache'


@dataclass
class ExecutionContext:
    actor: 'xrpc.actor.Actor'
    el: EventLoop
    chans: Dict[str, int]
    """Channels of `el` available in this context by these names"""
    chan_def: Optional[str] = None
    """Default channel to send messages from (as the client)"""

    origin: Optional[Origin] = None
    reply: Optional[Callable[[Any], None]] = None

    def exec(self, __origin, __fn: Callable, *args, **kwargs):
        is_ok = False
        r = None

        idx = context_push(self)

        t_start = time_now()

        log_tr_exec_in.debug('[%s %s] Name=%s %s %s %s %s', idx, __origin, __fn, is_ok, args, kwargs, r)

        try:
            r = __fn(*args, **kwargs)
            is_ok = True
            return r
        finally:
            t_end = time_now()
            secs = (t_start - t_end).total_seconds()

            if secs > 0.01:
                log_tr_exec_out.error('[%s %s], Name=%s (%s)', idx, __origin, __fn, secs)

            log_tr_exec_out.debug('[%s %s] Name=%s %s %s %s %s', idx, __origin, __fn, is_ok, args, kwargs, r)
            context_pop(idx)


def context_push(config: Optional[ExecutionContext]):
    if not hasattr(RUNTIME_TL, CTX_NAME):
        setattr(RUNTIME_TL, CTX_NAME, [])

    ctx = getattr(RUNTIME_TL, CTX_NAME)
    ctx.append(config)

    return len(ctx) - 1


def context_pop(idx: int):
    ctx = getattr(RUNTIME_TL, CTX_NAME)

    # this is required for calls that have returned early, but since have started processing an another executable
    new_ctx = ctx[:idx] + ctx[idx + 1:]

    setattr(RUNTIME_TL, CTX_NAME, new_ctx)


def context() -> ExecutionContext:
    return getattr(RUNTIME_TL, CTX_NAME)[-1]


def context_raw() -> Optional[List[ExecutionContext]]:
    try:
        return getattr(RUNTIME_TL, CTX_NAME)
    except AttributeError:
        return None


T = TypeVar('T')


def cache_get(obj: Type[T]) -> ServiceDefn:
    if not hasattr(RUNTIME_TL, CACHE_NAME):
        nd = {}
        setattr(RUNTIME_TL, CACHE_NAME, nd)

    cache = getattr(RUNTIME_TL, CACHE_NAME)

    if obj not in cache:
        cache[obj] = ServiceDefn.from_cls(obj)

    return cache[obj]


def origin(chan=None) -> Origin:
    ctx = context()
    return ctx.el.transport(ctx.chans[chan or ctx.chan_def]).origin


def sender() -> Origin:
    origin = context().origin
    assert origin is not None, 'sender() can only be called in RPC functions'

    return origin


def reply(ret: Any):
    reply = context().reply
    assert reply is not None, 'reply() can only be called in RPC functions'

    reply(ret)


def reset(name: Union[Callable, str], new_val: Any):
    if callable(name):
        name = name.__name__

    reg: xrpc.actor.RegularRunner = context().actor.get('regular')
    reg.reset(name, new_val)


TA = TypeVar('TA')
TB = TypeVar('TB')


def _masquerade(origin: str, orig: ServiceDefn, new: ServiceDefn, **map: str) -> str:
    """build an origin URL such that the orig has all of the mappings to new defined by map"""

    origin: ParseResult = urlparse(origin)

    prev_maps = {}

    if origin.query:
        prev_maps = {k: v for k, v in parse_qsl(origin.query)}

    r_args = {}

    for new_k, orig_k in map.items():

        assert new_k in new.rpcs, [new_k, new.rpcs]
        assert orig_k in orig.rpcs, [orig_k, orig.rpcs]

        # todo: check if the definitions are the same

        new_v = new.rpcs[new_k]
        orig_v = orig.rpcs[orig_k]

        if orig_k in prev_maps:
            orig_k = prev_maps[orig_k]

        assert new_v.res == orig_v.res, [new_v.res, orig_v.res]
        assert new_v.req == orig_v.req, [new_v.req, orig_v.req]

        r_args[new_k] = orig_k

    return urlunparse(origin._replace(query=urlencode(r_args)))


def masquerade(origin: str, orig: Type[TA], new: Type[TB], **map: str) -> str:
    """Make ``orig`` appear as new"""

    return _masquerade(origin, cache_get(orig), cache_get(new), **map)


def service(obj: Type[T], dest: Origin, conf: Optional[ClientConfig] = None, group=None, **kwargs) -> T:
    if conf is None:
        conf = ClientConfig(**kwargs)

    ctx = context()

    defn = cache_get(obj)

    group = ctx.chan_def if group is None else group

    assert group is not None, 'service() can only be called within execution context that references channels'

    trc('1').debug('%s %s %s', obj, dest, group)

    try:
        chan_idx = ctx.chans[group]
    except KeyError:
        raise ValueError(f'{group} not found in {ctx.chans.keys()}')
    else:
        return ServiceWrapper(defn, conf, ctx.el.transport(chan_idx), dest)



import logging
import socket
import time
from contextlib import contextmanager
from datetime import datetime
from functools import reduce
from itertools import count
from typing import Dict, List, Union, Optional, Callable, Tuple

from dataclasses import dataclass

from xrpc.net import RPCPacket
from xrpc.trace import log_tr_net_pkt_in, log_tr_net_pkt_err, log_tr_net_pkt_out
from xrpc.transport import Transport, \
    select_helper, Packet, RPCPacketRaw
from xrpc.util import time_now, _log_called_from, _log_traceback

EVPktMatcherFun = Callable[[RPCPacketRaw], bool]
EVPktHdlr = Callable[[RPCPacketRaw], None]
EVPollHdlr = Callable[[Optional[List[bool]]], None]
EVSleepSchedFun = Callable[[None], Optional[float]]
EVSleepTimerHdlr = Callable[[None], None]

EVExcHdlr = Callable[[BaseException], bool]


def default_exc_handler(exc: BaseException):
    _log_called_from(logging.getLogger(__name__), 'Ignoring uncaught')

    return True


def raise_exc_handler(exc: BaseException, should_log=True):
    if should_log:
        _log_called_from(logging.getLogger(__name__), 'Uncaught')

    return False


@dataclass
class ELPollEntry:
    # what happens if the event loop raises an exception ?
    # what happens if the event loop raises a TerminationException ?
    # who is supposed to catch them ?
    processor: EVPollHdlr
    exc: EVExcHdlr = default_exc_handler


@dataclass
class ELExcEntry:
    exc: EVExcHdlr = default_exc_handler


@dataclass
class ELPktEntry:
    matcher: EVPktMatcherFun
    processor: EVPktHdlr
    exc: EVExcHdlr = default_exc_handler


@dataclass
class ELWaitEntry:
    scheduler: EVSleepSchedFun
    processor: EVSleepTimerHdlr
    exc: EVExcHdlr = default_exc_handler


@contextmanager
def exc_handler(exc: EVExcHdlr):
    try:
        yield
    except BaseException as e:
        try:
            fe = not exc(e)
        except:
            logging.getLogger(__name__).exception('Could not handle the exception here')
            raise
        else:
            if fe:
                raise


class EventLoopEmpty(BaseException):
    pass


@dataclass
class ELSpinLockDetector:
    par: 'EventLoop'

    max_ctr: int = 3
    max_iter_secs: float = 0.01

    sleep_secs: float = 0.02

    flgs: Optional[List[bool]] = None
    flgs_ctr = 0
    time: Optional[datetime] = None

    def step(self, polled_flags):
        if any(polled_flags) and polled_flags == self.flgs:
            self.flgs_ctr += 1
        elif self.flgs is None:
            self.flgs = polled_flags
            self.flgs_ctr = 0
            self.time = time_now()
        else:
            self.flgs = polled_flags
            self.flgs_ctr = 0
            self.time = time_now()

        fa = any(self.flgs) and self.flgs_ctr > self.max_ctr
        vb = fa and ((time_now() - self.time).total_seconds() / self.flgs_ctr)
        fb = fa and vb < self.max_iter_secs

        if fa and fb:
            lgr = self.par.logger('spinlock')
            lgr.warning('%s %s %s', self.flgs_ctr, self.flgs, vb)
            lgr.warning('%s', self.par.nonflat_fds())
            lgr.warning('%s', self.par.transports)
            lgr.warning('%s', self.par.waits)

            _log_traceback(lgr, 'We are in a spinlock')

            time.sleep(self.sleep_secs)


def flatten(z):
    return reduce(lambda a, b: a + b, (x for _, x in z), [])


class EventLoop:
    def __init__(self):
        self.id_ctr = count()
        self.transports: Dict[int, Transport] = {}
        self.stacks: Dict[int, List[Union[ELPollEntry, ELPktEntry]]] = {}
        self.stacks_exc: Dict[int, List[ELExcEntry]] = {}
        self.waits: Dict[int, ELWaitEntry] = {}

    def logger(self, sn=None):
        name = __name__
        if sn:
            name = name + '.' + sn

        return logging.getLogger(name)

    def nonflat_fds(self) -> List[Tuple[int, List[socket.socket]]]:
        r = []
        for k, transport in sorted(self.transports.items()):
            r2 = []
            for y in transport.fds:
                r2.append(y)
            r.append((k, r2))
        return r

    def _is_raw(self, idx: int):
        stack = self.stacks[idx]

        return len(stack) and isinstance(stack[0], ELPktEntry)

    @property
    def max_waits(self) -> Dict[int, Optional[float]]:
        r: Dict[int, Optional[float]] = {}
        for k, v in self.waits.items():
            sched_item = v.scheduler()

            if sched_item is None:
                r[k] = None
                continue

            if sched_item < 0.:
                if -sched_item > 0.010:
                    self.logger('max_waits.e').warning('`%s` (%s < 0.)', v.scheduler, sched_item)
                sched_item = 0.
            r[k] = sched_item

        self.logger('max_waits.o').debug('%s', r)
        return r

    def loop(self, max_wait: Optional[float] = None):
        ts = time_now()

        def bin_none(a, b, fn=min):
            if a is None:
                return b
            elif b is None:
                return a
            else:
                return fn(a, b)

        elsld = ELSpinLockDetector(self)

        while True:
            running_for = (time_now() - ts).total_seconds()

            max_wait_left = None if max_wait is None else max(max_wait - running_for, 0)

            max_wait_owned: Optional[float] = None

            max_waits_curr = self.max_waits

            for k, x in max_waits_curr.items():
                max_wait_owned = bin_none(max_wait_owned, x)

                if x is not None and x <= 0.:
                    wait = self.waits.get(k)

                    if wait is None:
                        continue

                    with exc_handler(wait.exc):
                        wait.processor()

            max_wait_step = bin_none(max_wait_owned, max_wait_left)

            while True:
                if not len(self.waits) and not len(self.transports):
                    self.logger('ex').debug('Exit')
                    raise EventLoopEmpty()

                fds = self.nonflat_fds()

                # todo: SIGTERM is called here.
                try:
                    polled_flags = select_helper(flatten(fds), max_wait=max_wait_step)
                except ConnectionAbortedError:
                    self.logger('coabt.s').debug('%s', fds)
                    continue

                elsld.step(polled_flags)

                self.recv(fds, polled_flags)

                break

            if max_wait_left is not None and max_wait_left == 0:
                break

    def recv_connabt(self, idx: int, has_data, max_iter=5):
        for ctr in count():
            if idx not in self.transports:
                return

            if ctr >= max_iter:
                return

            transport = self.transports[idx]

            try:
                reader = transport.read(has_data)
                raw_packet = next(reader)
                yield raw_packet
            except StopIteration:
                return
            except Exception as e:
                exc_stack = self.stacks_exc[idx]

                for x in exc_stack[::-1]:
                    fl = x.exc(e)

                    if not fl:
                        raise
                    else:
                        return

                self.logger('coerr').debug('%s %s', e, transport)
                return

    def recv(self, nonflat_fds: List[Tuple[int, List[socket.socket]]], polled_flags: Optional[List[bool]] = None):
        # we have a set of transports
        # we've got a set of

        initial_polled_flags = polled_flags

        if polled_flags is None:
            polled_flags = [(k, [True for _ in fds]) for k, fds in nonflat_fds]
        else:
            polled_flags_nonflat = []
            idx = 0

            for k, fds in nonflat_fds:

                curr_polled_flags = []
                for _ in fds:
                    curr_polled_flags.append(polled_flags[idx])
                    idx += 1

                polled_flags_nonflat.append((k, curr_polled_flags))

            polled_flags = polled_flags_nonflat

        self.logger('recv.ep').debug('%s %s %s', nonflat_fds, polled_flags, initial_polled_flags)

        try:

            if not polled_flags:
                return []

            for idx, has_data in polled_flags:
                fa, fb = not any(has_data), idx not in self.transports

                if fa:
                    continue

                if fb:
                    continue

                transport = self.transports[idx]
                stack = self.stacks[idx]

                if len(stack) == 0:
                    self.logger('0st').warning('%s %s', transport, stack)
                    continue

                if len(stack) and isinstance(stack[-1], ELPollEntry):
                    with exc_handler(stack[-1].exc):

                        stack[-1].processor(has_data)
                else:
                    for raw_packet in self.recv_connabt(idx, has_data):
                        packet = RPCPacket.unpack(raw_packet.data)

                        log_tr_net_pkt_in.debug('%s %s %s', transport.origin, raw_packet.addr, packet)

                        raw_rpc_packet = RPCPacketRaw(raw_packet.addr, packet)

                        for st in stack[::-1]:
                            fm = st.matcher(raw_rpc_packet)
                            if fm:
                                with exc_handler(st.exc):
                                    st.processor(raw_rpc_packet)
                                break
                        else:
                            log_tr_net_pkt_err.error('[NO_MATCH] %s %s', raw_packet.addr, packet)
        finally:
            self.logger('recv.ex').debug('%s %s %s', nonflat_fds, polled_flags, initial_polled_flags)

    def send(self, idx: int, packet: RPCPacketRaw):
        chan = self.transports[idx]

        log_tr_net_pkt_out.debug('%s %s %s', chan.origin, packet.addr, packet.packet)

        raw_packet = Packet(packet.addr, packet.packet.pack())

        chan.send(raw_packet)

    def push(self, idx: int, entry: ELPktEntry):
        assert len(self.stacks[idx]) == 0 or all(isinstance(x, ELPktEntry) for x in self.stacks[idx]), self.stacks
        self.stacks[idx].append(entry)

    def push_exc(self, idx: int, entry: ELExcEntry):
        self.stacks_exc[idx].append(entry)

    def push_raw(self, idx: int, entry: ELPollEntry):
        assert len(self.stacks[idx]) == 0, self.stacks[idx]
        self.stacks[idx].append(entry)

    def pop(self, idx: int):
        assert idx in self.transports, (idx, self.transports)
        return self.stacks[idx].pop()

    def transport(self, idx: int) -> 'ELTransportRef':
        self.transports[idx]
        return ELTransportRef(self, idx)

    def wait(self, idx: int) -> 'ELWaitRef':
        assert idx in self.waits
        return ELWaitRef(self, idx)

    def wait_add(self, entry: ELWaitEntry) -> 'ELWaitRef':
        new_id = next(self.id_ctr)

        self.logger('wait.add').debug('%s %s %s', self, entry, new_id)

        self.waits[new_id] = entry

        return ELWaitRef(self, new_id)

    def wait_remove(self, idx: int):
        self.logger('wait.rm').debug('%s %s', self, idx)

        del self.waits[idx]

    def transport_add(self, transport: Transport) -> 'ELTransportRef':
        new_id = next(self.id_ctr)

        self.logger('tran.add').debug('%s %s %s', self, transport, new_id)

        self.transports[new_id] = transport
        self.stacks[new_id] = []
        self.stacks_exc[new_id] = []

        return self.transport(new_id)

    def transport_remove(self, idx: int) -> Transport:
        self.logger('tran.rm').debug('%s %s', self, idx)
        r = self.transports[idx]
        del self.transports[idx]
        del self.stacks[idx]
        del self.stacks_exc[idx]
        return r


@dataclass
class ELWaitRef:
    parent: EventLoop
    idx: int

    def remove(self):
        return self.parent.wait_remove(self.idx)


@dataclass
class ELTransportRef:
    parent: EventLoop
    idx: int

    def loop(self, max_wait: Optional[float] = None):
        return self.parent.loop(max_wait)

    def __getattr__(self, item):
        if item.startswith('__'):
            raise AttributeError(item)

        return getattr(self.parent.transports[self.idx], item)

    def is_alive(self) -> bool:
        try:
            self.parent.transport(self.idx)
            return True
        except KeyError:
            return False

    def send(self, packet: RPCPacketRaw):
        return self.parent.send(self.idx, packet)

    def push(self, entry: ELPktEntry):
        return self.parent.push(self.idx, entry)

    def push_exc(self, entry: ELExcEntry):
        return self.parent.push_exc(self.idx, entry)

    def push_raw(self, entry: ELPollEntry):
        return self.parent.push_raw(self.idx, entry)

    def pop(self):
        return self.parent.pop(self.idx)

    def remove(self):
        return self.parent.transport_remove(self.idx)



from typing import List

from xrpc.client import ClientConfig
from xrpc.dsl import rpc, signal
from xrpc.error import TerminationException
from xrpc.runtime import service, origin, sender


class Recursive:
    """Recursively call itself till exit"""

    @rpc()
    def ep(self, a: int) -> int:
        assert a > 0

        s = service(Recursive, origin(), ClientConfig(timeout_total=1., horz=False))

        if a == 1:
            return 0
        elif a > 1:
            return s.ep(a - 1)

    @signal()
    def exit(self):
        raise TerminationException()


class RecursiveA:
    """Call `RecursiveB` in order to return"""

    def __init__(self, url):
        self.url = url

    @rpc()
    def status(self) -> List[str]:
        return ['running', 'pending']

    @rpc()
    def poll(self, name: str = 'running') -> int:
        s = service(RecursiveB, self.url, ClientConfig(timeout_total=1., horz=False))

        return s.count_status(name)

    @signal()
    def exit(self):
        raise TerminationException()


class RecursiveC(RecursiveA):
    @rpc()
    def status(self) -> List[str]:
        raise ValueError()


class RecursiveB:
    """Make a call to the sender of the message"""

    @rpc()
    def count_status(self, name: str = 'running') -> int:
        s = service(RecursiveA, sender(), ClientConfig(timeout_total=1., horz=False))

        return len([x for x in s.status() if x == name])

    @signal()
    def exit(self):
        raise TerminationException()



import traceback
from contextlib import contextmanager

import pytz
import signal
from datetime import datetime
from typing import Union, Any


@contextmanager
def signal_context(signals=(signal.SIGINT, signal.SIGTERM), handler: Union[int, Any] = signal.SIG_IGN):
    prev_hdlrs = [signal.signal(s, handler) for s in signals]

    try:
        yield prev_hdlrs
    finally:
        for s, prev_hdlr in zip(signals, prev_hdlrs):
            signal.signal(s, prev_hdlr)


ISO8601 = '%Y-%m-%dT%H:%M:%S.%fZ'
ISO8601_SYMSAFE = '%Y%m%dT%H%M%S%fZ'


def time_now() -> datetime:
    return datetime.now(tz=pytz.utc)


def time_parse(v, format) -> datetime:
    return datetime.strptime(v, format).replace(tzinfo=pytz.utc)


def _build_callstack(ignore=1):
    assert ignore > 0

    INDENT = '  '

    callstack = '\n'.join([INDENT + line.strip() for line in traceback.format_stack()][:-ignore])

    return callstack


def _log_called_from(logger, pat='', *args):
    if len(pat):
        pat += '\n'

    logger.exception(pat + 'Called from\n%s\n', *args, _build_callstack())


def _log_traceback(logger, pat=''):
    if len(pat):
        pat += '\n'

    try:
        raise KeyError()
    except:
        logger.debug(pat + _build_callstack(ignore=2))



import base64
import logging
import multiprocessing
import os
import sys
import zlib
from contextlib import ExitStack, contextmanager

import dill
import signal
import subprocess
from dataclasses import dataclass
from datetime import timedelta
from time import sleep
from typing import Callable, Any, Tuple, Dict, List, Optional

from xrpc.util import signal_context, time_now, _build_callstack


@contextmanager
def cov():
    import coverage

    cov = None
    try:
        cov = coverage.process_startup()
        yield
    finally:
        if cov:
            cov.save()


def argv_encode(x: Any):
    return base64.b64encode(zlib.compress(dill.dumps(x), level=9)).decode()


def argv_decode(x: Any):
    return dill.loads(zlib.decompress(base64.b64decode(x)))


def wait_all(*waiting, max_wait=5.) -> List[int]:
    wait_till = time_now() + timedelta(seconds=max_wait)
    waiting = {i: v for i, v in enumerate(waiting)}

    r: Dict[int, int] = {}

    while wait_till > time_now() and len(waiting):
        for i, x in list(waiting.items()):
            try:
                r[i] = x.wait(0)

                del waiting[i]
            except multiprocessing.context.TimeoutError:
                pass
            except subprocess.TimeoutExpired:
                pass
        sleep(0.03)

    if len(waiting):
        raise TimeoutError(f'{waiting}')

    return [v for _, v in sorted(r.items())]


@dataclass
class PopenStruct:
    traceback: str
    fn: Callable
    args: Tuple[Any]
    kwargs: Dict[str, Any]


def popen_encode(fn, *args, **kwargs) -> List[str]:
    return [
        sys.executable,
        '-m',
        __name__,
        f'{fn.__module__}.{fn.__name__}',
        argv_encode(PopenStruct(
            _build_callstack(ignore=2),
            fn,
            args,
            kwargs
        )),
    ]


def popen(fn, *args, **kwargs) -> subprocess.Popen:
    """
    Please ensure you're not killing the process before it had started properly
    :param fn:
    :param args:
    :param kwargs:
    :return:
    """

    args = popen_encode(fn, *args, **kwargs)

    logging.getLogger(__name__).debug('Start %s', args)

    p = subprocess.Popen(args)
    return p


class PopenStackException(Exception):
    def __init__(self, abuser: Optional[subprocess.Popen] = None):
        super().__init__(abuser)

        self.abuser = abuser

    def __str__(self):
        return f'We have timed out and therefore killed all of the subprocesses (ABUSER: {self.abuser})'


class PopenStack:
    """This is mainly for GC purposes. We wait instead of SIGKILL immediately."""

    def __init__(self, timeout=None):
        self.stack: List[subprocess.Popen] = []
        self.timeout = timeout

    def add(self, x: subprocess.Popen):
        self.stack.append(x)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logging.getLogger(__name__).debug('Exit %s %s %s', exc_type, exc_val, exc_tb)
        if exc_type:
            for x in self.stack:
                x.send_signal(signal.SIGKILL)

            raise PopenStackException()
        else:
            should_kill = False
            item = None

            for x in self.stack:
                try:
                    code = x.wait(timeout=self.timeout)
                except subprocess.TimeoutExpired:
                    item = x
                    should_kill = True
                    break

            if should_kill:
                for x in self.stack:
                    x.send_signal(signal.SIGKILL)
                raise PopenStackException(item)


def _popen_defn() -> Optional[PopenStruct]:
    return argv_decode(sys.argv[2])


def popen_main():
    with ExitStack() as es:
        if os.environ.get('COVERAGE_PROCESS_START'):
            es.enter_context(cov())

        def popen_signal_handler(code, frame):
            if callable(prev_handler[code]):
                prev_handler[code](code, frame)

        codes = (signal.SIGINT, signal.SIGTERM)

        with signal_context(signals=codes, handler=popen_signal_handler) as prev_handlers:
            prev_handler = {k: v for k, v in zip(codes, prev_handlers)}

            defn = None

            try:
                defn: PopenStruct = _popen_defn()
            except:
                import traceback
                import pprint

                fmtd = pprint.pformat(sys.argv)
                print(defn.traceback, file=sys.stderr)
                traceback.print_exc(file=sys.stderr)
                sys.stderr.flush()
                raise ValueError(f'Cannot unpickle arguments, called with {fmtd}')

            defn.fn(*defn.args, **defn.kwargs)


if __name__ == '__main__':
    popen_main()



import logging

from xrpc.dsl import rpc, RPCType, signal
from xrpc.error import TerminationException

LIVELY = 'lively'


class Lively:
    @rpc(RPCType.Repliable, group=LIVELY)
    def is_alive(self) -> bool:
        return True


class Exceptional(Lively):
    """Handle exceptional socket conditions"""

    @rpc(exc=True)
    def ep(self, exc: Exception) -> bool:
        """Return False if the exception had not been handled gracefully"""
        if not isinstance(exc, ConnectionAbortedError):
            return False

        if len(exc.args) != 2:
            return False

        origin, reason = exc.args

        logging.getLogger(__name__).warning('Exited')

        return True

    @rpc(RPCType.Durable)
    def callme(self, x: int) -> int:
        return x + 1

    @signal()
    def exit(self):
        raise TerminationException()


class ExceptionalDropper(Exceptional):
    """Handle exceptional socket conditions"""

    @rpc(exc=True)
    def ep(self, exc: Exception) -> bool:
        pass


class ExceptionalClient(Lively):
    @rpc(exc=True)
    def ep(self, exc: ConnectionAbortedError) -> bool:
        return True

    @rpc(RPCType.Durable)
    def callme(self, x: int) -> int:
        return x + 1

    @signal()
    def exit(self):
        raise TerminationException()


class Client:
    @rpc(RPCType.Durable)
    def callme(self, x: int) -> int:
        return x + 1



import logging
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from typing import Optional, NamedTuple, Any, Type, ContextManager, TypeVar
from urllib.parse import urlparse, parse_qs, ParseResult, urlunparse

from dataclasses import dataclass, field

from xrpc.dsl import RPCType
from xrpc.error import TimeoutError, InvalidFingerprintError, HorizonPassedError, InternalError
from xrpc.loop import EventLoop, ELTransportRef, ELPktEntry, raise_exc_handler
from xrpc.net import RPCKey, RPCPacket, RPCReply, RPCPacketType
from xrpc.service import ServiceDefn
from xrpc.transport import Origin, RPCPacketRaw, Transport
from xrpc.util import time_now


class ClientConfig(NamedTuple):
    timeout_resend: float = 0.033
    timeout_total: Optional[float] = 3.5
    horz: bool = True
    """If ```HorizonPassedError``` is passed, restart the action"""


def dest_overrides(url):
    r: ParseResult = urlparse(url)
    qs = parse_qs(r.query) if r.query else {}
    r = r._replace(query=None)
    r: str = urlunparse(r)

    return {k: v[0] for k, v in qs.items()}, r


class ServiceWrapper:
    def __init__(self, defn: ServiceDefn, conf: ClientConfig, tran: ELTransportRef, dest: Origin):
        self.tran = tran
        self.overrides, self.dest = dest_overrides(dest)
        self.defn = defn
        self.conf = conf

        self._assert()

    def _assert(self):
        missing = []
        for x in self.overrides.keys():
            if x not in self.defn:
                missing.append(x)
        if len(missing):
            assert False, missing

    def __getattr__(self, item):
        f1 = item in self.defn
        alias = self.overrides.get(item)

        if f1 or alias:
            return CallWrapper(self, item, alias)
        else:
            logging.getLogger(__name__).error('%s', self.defn)
            raise AttributeError(item)


class RequestStopper(BaseException):
    def __init__(self, return_: Any, *args: object, **kwargs: object) -> None:
        self.return_ = return_
        super().__init__(return_, *args, **kwargs)


@dataclass
class RequestWrapper:
    par: ServiceWrapper
    name: str
    alias: Optional[str]
    key: RPCKey = field(default_factory=RPCKey)

    def __post_init__(self):
        self.call = self.par.defn[self.name]

    def __call__(self, *args, **kwargs):
        payload = self.par.defn.serde.serialize(self.call.req, [args, kwargs])

        packet = RPCPacket(self.key, RPCPacketType.Req, self.alias or self.name, payload)

        # the only difference between a client and a server is NONE.
        # the only issue would be the routing of the required packets to the required instances

        if self.call.conf.type == RPCType.Signalling:
            self.par.tran.send(RPCPacketRaw(self.par.dest, packet))
        elif self.call.conf.type in [RPCType.Repliable, RPCType.Durable]:
            time_started = time_now()

            self.par.tran.push(ELPktEntry(
                lambda x: x.packet.key == self.key and x.packet.type == RPCPacketType.Rep,
                self.process_packet,
                partial(raise_exc_handler, should_log=False),
            ))

            try:
                while True:
                    time_step = time_now()

                    dur_passed = time_step - time_started

                    if self.par.conf.timeout_total is not None and dur_passed.total_seconds() > self.par.conf.timeout_total:
                        raise TimeoutError()

                    self.par.tran.send(RPCPacketRaw(self.par.dest, packet))

                    self.par.tran.loop(max(0., self.par.conf.timeout_resend))
            except RequestStopper as e:
                return e.return_
            finally:
                # todo: if someone causes a TerminationException,

                # todo: if a server causes an EventLoopEmpty while doing this
                # todo: then the current transport is going to be destroyed before we reach this line

                if self.par.tran.is_alive():
                    self.par.tran.pop()
        else:
            raise NotImplementedError(self.call.conf.type)

    def process_packet(self, x: RPCPacketRaw):
        packet = x.packet

        assert packet.key == self.key, (packet, self.key)

        if packet.name == RPCReply.ok.value:
            return_ = self.par.defn.serde.deserialize(self.call.res, packet.payload)
            raise RequestStopper(return_)
        elif packet.name == RPCReply.fingerprint.value:
            raise InvalidFingerprintError(self.par.defn.serde.deserialize(Optional[str], packet.payload))
        elif packet.name == RPCReply.horizon.value:
            raise HorizonPassedError(self.par.defn.serde.deserialize(datetime, packet.payload))
        elif packet.name == RPCReply.internal.value:
            raise InternalError(packet.payload)
        else:
            raise NotImplementedError(packet.name)


@dataclass()
class CallWrapper:
    type: ServiceWrapper
    name: str
    alias: Optional[str]

    def __call__(self, *args, **kwargs):
        while True:
            try:
                return RequestWrapper(self.type, self.name, self.alias)(*args, **kwargs)
            except HorizonPassedError:
                if not self.type.conf.horz:
                    continue
                raise


def build_wrapper(pt: ServiceDefn, tran: ELTransportRef, dest: Origin, conf: ClientConfig = ClientConfig()):
    return ServiceWrapper(pt, conf, tran, dest)


class ClientTransportCircuitBreaker(BaseException):
    def __init__(self, pkt):
        self.pkt = pkt
        super().__init__(pkt)


T = TypeVar('T')


@contextmanager
def client_transport(
        rpc: Type[T],
        dest: str = 'udp://127.0.0.1:7483',
        conf: Optional[ClientConfig] = None,
        origin: str = 'udp://127.0.0.1',
        **kwargs
) -> ContextManager[T]:
    t = Transport.from_url(origin)

    if conf is None:
        conf = ClientConfig(**kwargs)

    def client_cb(p):
        raise ClientTransportCircuitBreaker(p)

    with t:
        ts = EventLoop()
        tref = ts.transport_add(t)
        # we may receive packets which are replies to things that are too late
        tref.push(ELPktEntry(lambda p: p.packet.type != RPCPacketType.Rep, client_cb, raise_exc_handler))
        pt = ServiceDefn.from_cls(rpc)
        r: T = build_wrapper(pt, tref, dest, conf=conf)

        yield r



import inspect
import sys

from argparse import ArgumentParser
from dataclasses import fields, dataclass, MISSING
from datetime import timedelta
from typing import Any, TypeVar, Type, Dict, List, Optional

from xrpc.generic import build_generic_context
from xrpc.logging import _dict_split_prefix
from xrpc.serde.types import is_union

T = TypeVar('T')


@dataclass
class ParsableConf:
    names: List[str]
    type: Optional[Any] = None
    action: Optional[str] = None


def is_list(t):
    t, _ = build_generic_context(t)

    if sys.version_info >= (3, 7):
        if hasattr(t, '__origin__'):
            return t.__origin__ is list
    if inspect.isclass(t):
        return issubclass(t, List)
    else:
        return False


def _guess_type_action(dest, type_, default) -> ParsableConf:
    action = 'store'

    if is_union(type_):
        args = type_.__args__

        if len(args) == 2 and args[-1] == type(None):
            type_ = args[0]
        else:
            assert False, type_

    if type_ is bool:
        type_ = None

        if default is True:
            action = 'store_false'
        else:
            action = 'store_true'
    elif type_ is timedelta:
        type_ = lambda x: timedelta(seconds=int(x))
    elif is_list(type_):
        type_, = type_.__args__
        action = 'append'
    return ParsableConf([dest], type_, action)


class Parsable:
    @classmethod
    def overrides(cls) -> Dict[str, ParsableConf]:
        return {}

    @classmethod
    def parser(cls, prefix: str, argparse: ArgumentParser):
        overrides = cls.overrides()

        for f in fields(cls):
            if f.default_factory is not MISSING:
                default = f.default_factory()
            else:
                default = f.default

            type_ = f.type

            name = f'{f.name}'

            if prefix:
                dest = f'{prefix}_{name}'
            else:
                dest = name

            if name in overrides:
                conf = overrides[name]
            else:
                conf = _guess_type_action('--' + dest, type_, default)

            help = None

            if hasattr(f, '__doc__'):
                help = f.__doc__

            if True:

                if help:
                    help += ' '
                else:
                    help = ''

                help += '(default: %(default)s)'

            type_action = {}
            if conf.type:
                type_action['type'] = conf.type

            if conf.action:
                type_action['action'] = conf.action

            if default is not MISSING:
                type_action['default'] = default

            argparse.add_argument(
                *conf.names,
                dest=dest,
                help=help,
                **type_action
            )

    @classmethod
    def from_parser(cls: Type[T], prefix, d, forget_other=True) -> T:
        filtered_fields, unfiltered_fields = _dict_split_prefix(d, prefix + '_' if prefix else '')

        field_set = {x.name for x in fields(cls)}

        filtered_fields = {k: v for k, v in filtered_fields.items() if k in field_set}
        unfiltered_fields = {k: v for k, v in filtered_fields.items() if k not in field_set}
        unfiltered_fields = {**unfiltered_fields, **unfiltered_fields}

        if forget_other:
            return cls(**filtered_fields)
        else:
            return unfiltered_fields, cls(**filtered_fields)



import contextlib
import logging
import os
import shutil
from datetime import timedelta, datetime
from functools import partial
from itertools import count
from tempfile import mkdtemp
from typing import List, Dict, Optional, Callable, Any, ContextManager

from dataclasses import dataclass

from xrpc.dict import RPCLogDict
from xrpc.dsl import RPCType, DEFAULT_GROUP
from xrpc.error import HorizonPassedError, InvalidFingerprintError, TerminationException
from xrpc.loop import EventLoop, ELTransportRef, ELPktEntry, ELWaitEntry, ELPollEntry, EventLoopEmpty, ELExcEntry
from xrpc.net import RPCPacket, RPCReply, RPCPacketType
from xrpc.runtime import ExecutionContext
from xrpc.serde.abstract import SerdeStruct
from xrpc.service import SocketIOEntrySet, RegularEntrySet, RPCEntrySet, SignalEntrySet
from xrpc.trace import log_tr_act, trc
from xrpc.transport import Transport, RPCPacketRaw, Transports, Packet
from xrpc.util import signal_context, time_now, _log_called_from, _log_traceback


class Terminating:
    def terminate(self):
        pass


class TerminatingHandler:
    actor: 'Actor'

    @contextlib.contextmanager
    def exc_ctx(self, reg_def):
        try:
            yield
        # except EventLoopEmpty:
        #    raise
        except TerminationException:
            self.actor.terminate('te1_' + repr(reg_def))
        except EventLoopEmpty:
            raise
        except BaseException:
            _log_called_from(logging.getLogger(f'{__name__}.{self.__class__.__name__}'), 'While %s', reg_def)
            self.actor.terminate('be' + repr(reg_def))
            raise

    def exception_handler(self, exc: BaseException):
        if isinstance(exc, TerminationException):
            self.actor.terminate('te2_' + repr(exc))

        return False


class RegularRunner(Terminating, TerminatingHandler):
    def __init__(
            self,
            actor: 'Actor',
            el: EventLoop,
            regulars: RegularEntrySet,
    ):
        # if we integrate max_waits into the RPCTransportStack

        self.actor = actor

        self.wait = el.wait_add(ELWaitEntry(self.max_wait, self.timeout, self.exception_handler))

        self.regulars = regulars

        self.states_regulars: Dict[str, Optional[datetime]] = {
            k: time_now() + timedelta(seconds=x.conf.initial) if x.conf.initial is not None else None for k, x in
            self.regulars.items()
        }

        # if we make sure that max_wait is called for every execution of select_
        # then we know if sleeping is available

    def terminate(self):
        self.wait.remove()

    def reset(self, name: str, new_val: Optional[float]):
        assert isinstance(new_val, (float, int, None.__class__)), new_val

        step_time = time_now()

        if new_val is not None:
            new_val = step_time + timedelta(seconds=new_val)

        self.states_regulars[name] = new_val

    @property
    def max_poll_regulars(self):
        step_time = time_now()

        max_poll_regulars = {k: (x - step_time).total_seconds() for k, x in self.states_regulars.items()
                             if x is not None}

        return max_poll_regulars

    def max_wait(self) -> Optional[float]:
        items = list(self.max_poll_regulars.values())
        return min(items) if len(items) else None

    def timeout(self):
        # a regular does not know about the transports available to it!

        max_poll_regulars = self.max_poll_regulars

        for name, reg_def in ((k, self.regulars[k]) for k, v in max_poll_regulars.items() if v is not None and v <= 0.):
            with self.exc_ctx(reg_def):
                self.states_regulars[name] = None
                x: float = self.actor.ctx().exec('reg', reg_def.fn)

                if x is not None and x < 0:
                    trc().error('%s (%s < 0)', name, x)

                if x is not None:
                    self.states_regulars[name] = time_now() + timedelta(seconds=x)
                else:
                    self.states_regulars[name] = None


class SocketIORunner(Terminating, TerminatingHandler):
    def __init__(
            self,
            actor: 'Actor',
            ts: EventLoop,
            sios: SocketIOEntrySet,
    ):
        self.actor = actor
        # these guys need to be able to provide the context in which we may change our sio mappings
        self.socketios = sios

        self.chans = Transports()

        self.chan = ts = ts.transport_add(self.chans)
        ts.push_raw(ELPollEntry(self.poll, self.exception_handler))

        for k, sio in self.socketios.items():
            with self.exc_ctx(sio):
                ret = self._assert_tran(sio.fn(None))
                self.chans[k] = ret

    def _assert_tran(self, tran):
        assert isinstance(tran, Transport), tran
        return tran

    def terminate(self):
        self.chan.remove()

    def poll(self, polled_flags: Optional[List[bool]] = None):
        kv = self.chans.poll_helper(polled_flags)

        for key, ready_flags in kv.items():
            if not any(ready_flags):
                continue

            sio_def = self.socketios[key]

            with self.exc_ctx(sio_def):
                r = self.actor.ctx().exec('sio', sio_def.fn, ready_flags)

                r = self._assert_tran(r)

                self.chans[key] = r


SignalHdlr = Callable[[int, int], bool]


@dataclass
class SigHdlrEntry:
    act: int
    sigh: SignalHdlr


def signal_handler(sig, frame, self: 'SignalRunner' = None):
    self.signal_handler(sig, frame)


@dataclass
class SignalRunnerRef(Terminating):
    par: 'SignalRunner'
    idx: int

    def __getattr__(self, item):
        return getattr(self.par, item)

    def terminate(self):
        self.par.remove(self.idx)


class LoggingActor:
    def logger(self, sn=None):
        name = self.__class__.__name__

        if sn:
            name += '.' + sn
        return log_tr_act(name)


class SignalRunner(LoggingActor, TerminatingHandler):
    # remove an actor whenever any of it's signal handlers have returned true
    # need to dynamically manage the state of the current handlers.

    # an actor terminates whenever the call had been made to terminate it
    # a signal runner terminates whenever it's empty
    def __init__(
            self,
            el: EventLoop,
    ):
        self.el = el
        self.idx_ctr = count()
        self.acts: Dict[int, Actor] = {}
        self.ssighs: Dict[int, List[SigHdlrEntry]] = {}

        self.ces: Dict[int, ContextManager] = {}

        self.has_transport = False
        # we use that to wake up our running thread

    def establish_transport(self):

        self.has_transport = True

        self.path_temp = mkdtemp()

        self.path_unix = 'unix://' + os.path.join(self.path_temp, 'sock.sock')
        self.tran = Transport.from_url(self.path_unix + '#bind')
        self.tran_ref = self.el.transport_add(self.tran)
        self.tran_ref.push_raw(ELPollEntry(
            self.handle_packet,
        ))

        self.logger('tran.est').debug('%s', self.path_unix)

    def _fork_transport_patch(self):
        # unsafe

        if self.has_transport:
            self.has_transport = None
            self.path_temp = None
            self.tran_ref.remove()
            self.tran = None

            self.establish_transport()

    def demolish_transport(self):
        self.tran_ref.remove()
        self.tran.close()
        shutil.rmtree(self.path_temp)

        self.has_transport = False

        self.logger('tran.dem').debug('%s', self.path_unix)

    def add(self, act: 'Actor', signals: Dict[int, List[SignalHdlr]]) -> SignalRunnerRef:
        self.logger('act.add').info('%s %s', act, signals)

        new_idx = next(self.idx_ctr)
        self.acts[new_idx] = act

        for k, sighs in signals.items():
            if k not in self.ssighs:
                self._new_hdlr(k)
            for sigh in sighs:
                self.ssighs[k].append(SigHdlrEntry(new_idx, sigh))

        if len(self.ssighs) and not self.has_transport:
            self.establish_transport()

        return SignalRunnerRef(self, new_idx)

    def remove(self, act: int):
        self.logger('act.rm').info('%s %s %s', act, self.ssighs, self.has_transport)

        del self.acts[act]

        for k in list(self.ssighs.keys()):
            sighs = self.ssighs[k]

            try:
                sigh_idx = [x.act for x in sighs].index(act)
            except ValueError:
                continue
            else:
                sighs = sighs[:sigh_idx] + sighs[sigh_idx + 1:]

            if len(sighs) == 0:
                self._del_hdlr(k)
            else:
                self.ssighs[k] = sighs
        if len(self.ssighs) == 0 and self.has_transport:
            self.demolish_transport()

    def _new_hdlr(self, code: int):
        self.logger('hdlr.add').debug('%s', code)

        assert code not in self.ssighs, (code, self.ssighs)
        self.ces[code] = signal_context(signals=(code,), handler=partial(signal_handler, self=self))
        self.ces[code].__enter__()
        self.ssighs[code] = []

    def _del_hdlr(self, code: int):
        self.logger('hdlr.rm').debug('%s', code)
        del self.ssighs[code]
        try:
            self.ces[code].__exit__(None, None, None)
        except BaseException:
            self.logger('hdlr.rm').exception('%s', code)
            raise

    def handle_packet(self, flags: Optional[List[bool]]):
        while True:
            try:
                for packet in self.tran.read(flags):
                    code = int(packet.data.decode())

                    self.handle_actor(code)
            except ConnectionAbortedError:
                continue
            else:
                return

    def handle_actor(self, sig):
        for sighe in self.ssighs[sig]:
            if sighe.act not in self.acts:
                continue

            act = self.acts[sighe.act]

            try:
                act.ctx().exec('sig', sighe.sigh)
            except TerminationException:
                act.terminate('sigte')
            except:
                logging.getLogger(__name__ + '.' + self.__class__.__name__).exception('%s', sig)

    def signal_handler(self, sig, frame):
        self.logger('hdlr.x').info('%s %s', sig, frame)

        with Transport.from_url(self.path_unix) as t:
            sig = str(sig).encode()
            t.send(Packet(self.path_unix, sig))


@dataclass(eq=True)
class HandlingMarker:
    pass


HANDLING = HandlingMarker()


class RPCGroupRunner(Terminating, TerminatingHandler):
    def __init__(
            self,
            actor: 'Actor',
            group: str,
            chan: ELTransportRef,
            serde: SerdeStruct,
            rpcs: RPCEntrySet,
            horizon_each=60.
    ):
        self.actor = actor
        self.horizon_each = horizon_each

        self.log_dict = RPCLogDict(time_now())

        self.group = group
        self.chan = chan
        self.wait = actor.el.wait_add(ELWaitEntry(self.max_wait, self.timeout))

        self.chan.push(ELPktEntry(
            lambda x: x.packet.type == RPCPacketType.Req,
            self.packet_receive,
            self.exception_handler
        ))

        self.serde = serde

        self.rpcs: RPCEntrySet = {k: v for k, v in rpcs.items() if not v.conf.exc}

        self.rpcs_exc: RPCEntrySet = {k: v for k, v in rpcs.items() if v.conf.exc}

        self.chan.push_exc(ELExcEntry(self.handle_exc))

    def handle_exc(self, exc: BaseException) -> bool:
        ctx = self.actor.ctx()
        for k, v in self.rpcs_exc.items():
            try:
                if ctx.exec('tran_exc', v.fn, exc):
                    return True
            except TerminationException:
                self.actor.terminate('te')
                return True
        else:
            return False

    def terminate(self):
        self.wait.remove()

    def get_horizon_wait(self) -> float:
        step_time = time_now()

        next_horizon: datetime = self.log_dict.horizon + timedelta(seconds=self.horizon_each) * 2

        max_poll_horizon = (next_horizon - step_time).total_seconds()

        return max_poll_horizon

    def max_wait(self) -> Optional[float]:
        return self.get_horizon_wait()

    def timeout(self):
        self.log_dict.set_horizon(time_now() - timedelta(seconds=self.horizon_each))

    def _get_rpc(self, name):
        if name not in self.rpcs:
            _log_traceback(trc('1'))
            trc('0').error('%s %s %s %s', self.chan.origin, self.group, name, self.rpcs.keys())
            raise InvalidFingerprintError('name')

        rpc_def = self.rpcs[name]

        return rpc_def

    def has_returned(self, key):
        return key in self.log_dict and self.log_dict[key] != HANDLING

    def packet_reply(self, raw_packet: RPCPacketRaw, p: RPCPacket, ret: Any):
        rpc_def = self._get_rpc(p.name)

        assert rpc_def.conf.type == RPCType.Repliable, p.name

        has_returned = self.has_returned(p.key)
        assert not has_returned

        ret_payload = self.serde.serialize(rpc_def.res, ret)

        self.log_dict[p.key] = ret_payload

        rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.ok.value, ret_payload)

        self.chan.send(RPCPacketRaw(raw_packet.addr, rp))

    def packet_handle(self, packet: RPCPacket, raw_packet: RPCPacketRaw, args, kwargs):
        rpc_def = self._get_rpc(packet.name)

        # try:
        # some RPCs may return earlier than exiting themselves
        # we make sure the RPC can not return the value twice

        has_returned = self.has_returned(packet.key)

        ctx = self.actor.ctx(
            chan_def=self.group,
            origin=raw_packet.addr,
            reply=partial(self.packet_reply, raw_packet, packet)
        )

        try:
            ret = ctx.exec('call', rpc_def.fn, *args, **kwargs)
        except TerminationException:
            raise

        if rpc_def.conf.type == RPCType.Repliable:
            if not has_returned:
                self.packet_reply(raw_packet, packet, ret)
            elif has_returned and ret is None:
                pass
            elif has_returned and ret is not None:
                raise ValueError(self.log_dict[packet.key])
            else:
                raise NotImplementedError()

    def packet_receive(self, raw_packet: RPCPacketRaw):
        pkt = raw_packet.packet

        def reply_now(a: str, b: Any):
            rp = RPCPacket(pkt.key, RPCPacketType.Rep, a, b)

            self.chan.send(RPCPacketRaw(raw_packet.addr, rp))

        try:
            rpc_def = self._get_rpc(pkt.name)
        except InvalidFingerprintError as e:
            reply_now(RPCReply.fingerprint.value, self.serde.serialize(Optional[str], e.reason))
            return

        try:
            if pkt.key in self.log_dict:
                rep = self.log_dict[pkt.key]

                if rep == HANDLING:
                    # todo short-circuit packets that we have received (maybe acknowledge them even)
                    return

                if rpc_def.conf.type != RPCType.Signalling:
                    reply_now(RPCReply.ok.value, rep)

                return
        except HorizonPassedError as e:
            reply_now(RPCReply.horizon.value, self.serde.serialize(datetime, e.when))
            return

        try:
            args, kwargs = self.serde.deserialize(rpc_def.req, pkt.payload)
        except Exception as e:
            # could not deserialize the payload correctly
            logging.exception(f'Failed to deserialize packet from {raw_packet.addr}')
            reply_now(RPCReply.fingerprint.value, self.serde.serialize(Optional[str], 'args'))
            return

        self.log_dict[pkt.key] = HANDLING

        if rpc_def.conf.type == RPCType.Durable:
            reply_now(RPCReply.ok.value, None)
            self.log_dict[pkt.key] = None
        elif rpc_def.conf.type == RPCType.Signalling:
            self.log_dict[pkt.key] = None

        try:
            self.packet_handle(pkt, raw_packet, args, kwargs)
        except Exception as e:
            # reply_now(RPCReply.internal.value, str(e))
            _log_called_from(logging.getLogger(__name__), 'While receiving the payload [%s %s %s] %s', pkt, args,
                             kwargs, rpc_def)
            self.actor.terminate('ie_' + repr(e))
            raise
        except TerminationException as e:
            if rpc_def.conf.type == RPCType.Repliable:
                reply_now(RPCReply.ok.value, self.serde.serialize(rpc_def.res, None))
            self.actor.terminate('te4_' + repr(e))


class Actor(Terminating, LoggingActor):
    def __init__(self, el: EventLoop, name: Optional[str] = None):
        self.el = el

        self.name = name

        self.idx_ctr = count()
        self.terms: Dict[int, Terminating] = {}
        self.names_terms: Dict[str, int] = {}
        self.terms_names: Dict[int, str] = {}
        self.chans: Dict[str, int] = {}

    def add_transport(self, group: str, url: str) -> ELTransportRef:
        tran = Transport.from_url(url)

        tref = self.el.transport_add(tran)

        self.chans[group] = tref.idx

        return tref

    def get(self, name) -> Terminating:
        return self.terms[self.names_terms[name]]

    def add(self, item: Terminating, name: Optional[str] = None) -> int:
        new_idx = next(self.idx_ctr)
        self.terms[new_idx] = item

        if name is not None:
            assert name not in self.names_terms, (self.names_terms, name)
            self.names_terms[name] = new_idx
            self.terms_names[new_idx] = name

        return new_idx

    def remove(self, idx: int):
        del self.terms[idx]

        if idx in self.terms_names:
            name = self.terms_names[idx]
            del self.terms_names[idx]
            del self.names_terms[name]

    def terminate(self, why=None):
        self.logger('term').info('Terminating %s %s %s', self.name, why, self)

        for k, v in list(self.chans.items()):
            del self.chans[k]

            tran = self.el.transport(v).remove()

            self.logger('term.tran').info('Closing %s %s %s', self.name, k, v)
            tran.close()

        for k, v in list(self.terms.items()):
            del self.terms[k]
            self.logger('term.act').info('Terminating %s %s %s', self.name, why, v)
            v.terminate()

    def ctx(self, **kwargs) -> ExecutionContext:
        if 'chan_def' not in kwargs:
            if DEFAULT_GROUP in self.chans:
                kwargs['chan_def'] = DEFAULT_GROUP

        return ExecutionContext(
            actor=self,
            el=self.el,
            chans=self.chans,
            **kwargs
        )


def actor_create(
        el: EventLoop,
        sr: SignalRunner,
        cls, cls_inst, bind_urls: Dict[str, str],
        horizon_each=60.,
        name: Optional[str] = None
) -> 'Actor':
    if isinstance(bind_urls, list):
        assert len(bind_urls) == 1, bind_urls

        bind_urls = {DEFAULT_GROUP: bind_urls[0]}

    rpcs = RPCEntrySet.from_cls(cls).bind(cls_inst)
    regs = RegularEntrySet.from_cls(cls).bind(cls_inst)

    sios = SocketIOEntrySet.from_cls(cls).bind(cls_inst)
    sigs = SignalEntrySet.from_cls(cls).bind(cls_inst)

    tran_grps = sorted(bind_urls.keys())
    rpc_grps = sorted(rpcs.groups())

    rpc_requested = set(rpc_grps) - set(tran_grps)

    assert len(rpc_requested) == 0, rpc_requested

    act = Actor(el, name=name)

    for tran_grp in tran_grps:
        trc('tran').debug('%s %s', tran_grp, bind_urls[tran_grp])

        tran_ref = act.add_transport(tran_grp, bind_urls[tran_grp])

        if tran_grp not in rpc_grps:
            continue

        srpcs = rpcs.by_group(tran_grp)

        act.add(
            RPCGroupRunner(
                act,
                tran_grp,
                tran_ref,
                srpcs.serde,
                srpcs,
                horizon_each=horizon_each
            )
        )
    regr = RegularRunner(act, el, regs)
    sior = SocketIORunner(act, el, sios)

    act.add(regr, 'regular')
    act.add(sior)

    act.add(sr.add(act, sigs.to_signal_map()), 'signal')

    return act


def run_server(cls, cls_inst, bind_urls: Dict[str, str], horizon_each=60., actor_name=None):
    el = EventLoop()
    sr = SignalRunner(el)

    if actor_name is None:
        if hasattr(cls, '__name__'):
            actor_name = cls.__name__
        else:
            actor_name = str(cls)

    act = actor_create(el, sr, cls, cls_inst, bind_urls, horizon_each, name=actor_name)

    try:
        el.loop()
    except EventLoopEmpty:
        pass
    except:
        act.terminate('exit')
        raise



import gc
import inspect
import logging

from typing import Optional

log_tr = logging.getLogger('xrpc.tr')

log_tr_exec_in = log_tr.getChild('x.i')
log_tr_exec_out = log_tr.getChild('x.o')

log_tr_act = lambda x: log_tr.getChild('a').getChild(x)

log_tr_net_sel_in = log_tr.getChild('n.s.i')
log_tr_net_sel_err = log_tr.getChild('n.s.e')

log_tr_net_obj_in = log_tr.getChild('n.o.i')
log_tr_net_obj_out = log_tr.getChild('n.o.i')

log_tr_net_raw_in = log_tr.getChild('n.r.i')
log_tr_net_raw_out = log_tr.getChild('n.r.o')
log_tr_net_raw_err = log_tr.getChild('n.r.e')

log_tr_net_pkt_in = log_tr.getChild('n.p.i')
log_tr_net_pkt_out = log_tr.getChild('n.p.o')
log_tr_net_pkt_err = log_tr.getChild('n.p.e')

log_tr_net_meta_in = log_tr.getChild('n.m.i')
log_tr_net_meta_out = log_tr.getChild('n.m.o')


def trc(postfix: Optional[str] = None, *, depth=1) -> logging.Logger:
    """
    Automatically generate a logger from the calling function

    :param postfix: append another logger name on top this
    :param depth: depth of the call stack at which to capture the caller name
    :return: instance of a logger with a correct path to a current caller

    """
    x = inspect.stack()[depth]

    mod = inspect.getmodule(x.frame)

    code = x[0].f_code

    try:
        func = next(obj for obj in gc.get_referrers(code) if inspect.isfunction(obj))
        parts = (mod.__name__, f'{func.__qualname__}:{x.lineno}')
    except StopIteration:
        parts = (f'{code.co_filename}:{x.lineno}',)

    if postfix:
        parts += (postfix,)

    logger_name = '.'.join(parts)

    return logging.getLogger(logger_name)



import socket

import datetime
from dataclasses import dataclass, replace
from functools import partial
from typing import Any, Optional, Type, Callable, Dict, List

from xrpc.const import SERVER_SERDE_INST
from xrpc.dsl import rpc, regular, socketio, signal, DEFAULT_GROUP
from xrpc.generic import build_generic_context
from xrpc.serde.abstract import SerdeStruct, SerdeSet
from xrpc.serde.types import CallableArgsWrapper, CallableRetWrapper
from xrpc.transform import get_rpc, get_regular, get_socketio, get_signal


class Bindable(dict):
    def bind(self, obj: Any):
        return self.__class__({k: replace(v, fn=partial(v.fn, obj)) for k, v in self.items()})


@dataclass
class RPCEntry:
    name: str
    fn: Callable
    conf: rpc
    req: Type[Any]
    res: Type[Any]


class RPCEntrySet(Bindable, Dict[str, RPCEntry]):
    def groups(self):
        r = set()

        for v in self.values():
            r.add(v.conf.group)

        return sorted(list(r))

    def by_group(self, name=DEFAULT_GROUP) -> 'RPCEntrySet':
        return RPCEntrySet({
            k: v for k, v in self.items() if v.conf.group == name
        })

    @property
    def serde(self) -> SerdeStruct:
        ss = SerdeSet.walk(SERVER_SERDE_INST, datetime.datetime)
        ss = ss | SerdeSet.walk(SERVER_SERDE_INST, Optional[str])

        for k, v in self.items():
            if v.conf.exc:
                continue
            ssreq = SerdeSet.walk(SERVER_SERDE_INST, v.req)
            ssres = SerdeSet.walk(SERVER_SERDE_INST, v.res)

            ss = ss | ssreq | ssres

        return ss.struct(SERVER_SERDE_INST)

    @classmethod
    def from_cls(cls, type_):
        type_, ctx = build_generic_context(type_)

        rpcs = get_rpc(type_)
        rpcs_return = RPCEntrySet()

        for rpc_name, rpc_def in rpcs.items():
            fa = CallableArgsWrapper.from_func_cls(type_, rpc_def.fn, )
            fb = CallableRetWrapper.from_func_cls(type_, rpc_def.fn, )

            if rpc_def.conf.exc:
                retannot = fb.spec.annotations.get('return')
                assert retannot == bool, (rpc_def.fn, retannot)

            rpcs_return[rpc_name] = RPCEntry(
                rpc_name, rpc_def.fn, rpc_def.conf, fa, fb
            )

        return rpcs_return


@dataclass
class RegularEntry:
    name: str
    fn: Callable
    conf: regular


class RegularEntrySet(Bindable, Dict[str, RegularEntry]):
    @classmethod
    def from_cls(cls, type_):
        regulars = get_regular(type_)

        r = cls()
        for n, defn in regulars.items():
            r[n] = RegularEntry(
                n,
                defn.fn,
                defn.conf,
            )
        return r


@dataclass
class SocketIOEntry:
    name: str
    fn: Callable[[Optional[List[bool]]], List[socket.socket]]
    conf: socketio


class SocketIOEntrySet(Bindable, Dict[str, SocketIOEntry]):
    @classmethod
    def from_cls(cls, type_):
        sios = get_socketio(type_)

        r = cls()
        for n, (conf, fn) in sios.items():
            r[n] = SocketIOEntry(
                n,
                fn,
                conf
            )
        return r


@dataclass
class SignalEntry:
    name: str
    fn: Callable
    conf: signal


class SignalEntrySet(Bindable, Dict[str, SignalEntry]):
    def to_signal_map(self) -> Dict[int, List[Callable]]:
        r = {}

        for k, v in self.items():
            for code in v.conf.codes:
                if code not in r:
                    r[code] = []
                r[code].append(v.fn)
        return r

    @classmethod
    def from_cls(cls, type_):
        sios = get_signal(type_)

        r = cls()
        for n, (conf, fn) in sios.items():
            r[n] = SignalEntry(
                n,
                fn,
                conf
            )
        return r


@dataclass
class ServiceDefn:
    serde: SerdeStruct
    rpcs: RPCEntrySet

    def __contains__(self, item):
        return item in self.rpcs

    def keys(self):
        return self.rpcs.keys()

    def __getitem__(self, item) -> RPCEntry:
        return self.rpcs[item]

    def bind(self, obj) -> 'ServiceDefn':
        return ServiceDefn(self.serde, self.rpcs.bind(obj))

    @classmethod
    def from_cls(cls, type_):
        x = RPCEntrySet.from_cls(type_)

        return ServiceDefn(x.serde, x)
